#ifndef PR_PRIMER_H
#define PR_PRIMER_H

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include "dpal.h"

/* The following is somewhat experimental */
#define PR_INFINITE_POSITION_PENALTY -1.0
#define PR_DEFAULT_OUTSIDE_PENALTY    0.0
#define PR_DEFAULT_INSIDE_PENALTY     PR_INFINITE_POSITION_PENALTY


/* Maxima needed for interface data structures. */
#define PR_MAX_INTERVAL_ARRAY 200 /* 
				   * Maximum number of input intervals
				   * supported; used for targets, excluded
				   * regions, product-size intervals, etc.
				   */


#define PR_ALIGN_SCORE_PRECISION 100.0
#define PR_MAX_LIBRARY_WEIGHT 100.0

#define TRIMMED_SEQ_LEN(X) ((X)->incl_l)

#define MACRO_STRING(X) #X
/* pr_progam_name must be set in main(). */
#define PR_ASSERT(COND)                                  \
if (!(COND)) {                                           \
    fprintf(stderr, "%s:%s:%d, assertion (%s) failed\n", \
	   pr_program_name, __FILE__, __LINE__,          \
	   MACRO_STRING(COND));                          \
    abort();                                             \
}

typedef struct rep_sim {
  char *name;      /* Name of the sequence from given file in fasta
		    * format with maximum similarity to the oligo.
		    */
  short min;       /* 
                    * The minimum score in slot 'score' (below).
                    * (Used when the objective function involves
                    * minimization of mispriming possibilities.)
                    */
  short max;       /* The maximum score in slot 'score' (below). */
  short *score;    /* 
                    * Array of similarity (i.e. false-priming) scores,
                    * one for each entry in the 'repeat_lib' slot
                    * of the primargs struct. 
                    */
} rep_sim;

typedef struct primrec {
  rep_sim repeat_sim; /* Name of the sequence from given file in fasta
		       * format with maximum similarity to the oligo
		       * and corresponding alignment score.
		       */
  double temp;     /* 
		    * The oligo melting temperature calculated for the
		    * primer.
		    */
  double gc_content;
  double position_penalty; 
                  /*
                   * Penalty for distance from "ideal" position as specified
	           * by inside_penalty and outside_penalty.
                   */
  double quality;  /* Part of objective function due to this primer. */
  double end_stability; /* Delta G of disription of 5 3' bases. */
  int    start;    /* The 0-based index of the leftmost base of the primer. */
  int    seq_quality; /* Minimum quality score of bases included. */   
  short  self_any; /* Self complementarity as local alignment * 100. */
  short  self_end; /* Self complementarity at 3' end * 100. */
  char   target;   /* 
		    * 0 if this primer does not overlap any target, 1 if it
		    * does.
		    */
  char   excl;     /* 
		    * 0 if does not overlap any excluded region, 1 if it
		    * does.
		    */
  char   ok;       /* 
		    * -1 if is not acceptable, 0 if it is not known yet, 1 if
		    * ok.
		    */
  char   length;   /* Length of the oligo. */
  char   num_ns;   /* Number of Ns in the oligo. */
  char   position_penalty_infinite; 
                   /* Non-0 if the position penalty is infinite. */
} primer_rec;

/* 
 * The structure for a pair of primers. (So that we can have a function which
 * returns a pair of primers.)
 */
typedef struct primpair {
  double pair_quality;
  double compl_measure; /* 
			 * A measure of self-complementarity of left and right
			 * * primers in the pair, as well as complementarity *
			 * between left and right primers.  The function
			 * choice * returns pairs with the minimal value for
			 * this field * when 2 pairs have the same
			 * pair_quality.
			 */
  double diff_tm;       /* Absolute value of the difference between melting
			 * temperatures for left and right primers. 
			 */
  int    compl_any;     /* 
			 * Local complementarity score between left and right
			 * primers (* 100).
			 */
  int    compl_end;     /* 
		         * 3'-anchored global complementatory score between *
		         * left and right primers (* 100).
			 */

  short  repeat_sim;    /* Maximum total similarity of both primers to the
			 * sequence from given file in fasta format.
			 */
  primer_rec *left;     /* Left primer. */
  primer_rec *right;    /* Right primer. */
  primer_rec *intl;     /* Internal oligo. */

  int    product_size;  /* Product size. */
  int    target;        /* 
			 * 1 if there is a target between the right and left
			 * primers.
			 */
  char   *rep_name;
} primer_pair;

typedef struct pair_array_t {
    int         storage_size;
    int         num_pairs;
    primer_pair *pairs;
} pair_array_t;

typedef int interval_array_t[PR_MAX_INTERVAL_ARRAY][2];

/* pr_append_str is an append-only string ADT. */
typedef struct pr_append_str {
    int storage_size;
    char *data;
} pr_append_str;

/* The seq_lib struct represents a library of sequences. */
typedef struct seq_lib {
    char **names;         /* An array of sequence names. */
    char **seqs;          /* An array of sequences. */
    char **rev_compl_seqs;/* An array of reversed-complemented sequences. */
    double *weight;       /* An array of weights. */
    char   *repeat_file;  /* The path of the file containing the library. */
    pr_append_str error;  /* Global error message if any.  */
    pr_append_str warning;/* Warning message. */
    int seq_num;          /* The number of names, sequences, and weights. */
} seq_lib;

/* 
 * Arguments to the primer program as a whole.  Values for these arguments are
 * retained _across_ different input records.  (These are the so-called
 * "Global" arguments in the documentation.)
 */
typedef struct primargs {
    int    pr_min[PR_MAX_INTERVAL_ARRAY]; /* Minimum product sizes. */
    int    pr_max[PR_MAX_INTERVAL_ARRAY]; /* Maximum product sizes. */
    seq_lib repeat_lib;  /* Library of sequences to avoid. */

    pr_append_str glob_err;

    double opt_tm;
    double min_tm;
    double max_tm;
    double max_diff_tm;
    double max_gc;
    double min_gc;
    double salt_conc;
    double dna_conc;

    double outside_penalty; /* Multiply this value times the number of NTs
                             * from the 3' end to the the (unique) target to
                             * get the 'position penalty'.
                             * Meaningless if there are multiple targets
                             * or if the primer cannot be part of a pair
                             * that spans the target.
			     */

    double inside_penalty;  /* Multiply this value times the number of NT
                             * positions by which the primer overlaps
                             * the (unique) target to the 'position penalty'.
                             * Meaningless if there are multiple targets
                             * or if the primer cannot be part of a pair
                             * that spans the target.
                             */

    double max_end_stability;
                            /* The maximum value allowed for the delta
                             * G of disruption for the 5 3' bases of
                             * a primer.
                             */
    int    num_intervals;         /* 
				   * Number of product size intervals
				   * (i.e. number of elements in pr_min and
				   * pr_max)
                                   */
    int    num_ns_accepted;
    int    pick_internal_oligo;   /* 0 if don't pick. */
    int    compl_flag;            /* 0 if complementarity is not. */
    int    file_flag;
    int    explain_flag;
    int    primer_opt_size;
    int    primer_min_size;
    int    primer_max_size;

    int    gc_clamp;              /* Required number of GCs at *3' end. */
    short  repeat_compl;   /* 
			    * Acceptable complementarity with repeat
			    * sequences.
			    */

    short  pair_repeat_compl;
    int    liberal_base;   /* 
			    * If non-0 then turn characters other than
			    * [ATGCNatgcn] into N.
			    */

    int    max_poly_x;      /* 
			     * Maximum length of mononucleotide sequence in an
			     * oligo.
			     */

    int    first_base_index;  /* 
			       * The index of the first base in the input
			       * sequence.  This parameter is ignored within
			       * pr_choice; pr_choice's caller must assure that
			       * all indexes are 0-based.  However, this
			       * parameter should used by output routines to
			       * adjust base indexes.
			       */
    int    num_return; /* The number of best primer pairs to return. */
    int    min_quality;       /* Minimum quality permitted for oligo sequence.*/
    int    min_end_quality;   /* Minimum quality permitted at 3' end. */
    int    quality_range_min;
    int    quality_range_max;

    short  self_any;
    short  self_end;
    short  pair_compl_any;
    short  pair_compl_end;

} primer_args;

typedef struct pair_stats {
    int considered;         /* Total number of pairs or triples tested.      */
    int product;            /* Pairs providing incorrect product size.       */
    int target;             /* Pairs without any target between primers.     */
    int temp_diff;          /* Melting temperature difference too high.      */
    int compl_any;          /* Pairwise complementarity larger than allowed. */
    int compl_end;          /* The same for 3' end complementarity.          */
    int internal;           /* Internal oligo was not found.                 */
    int repeat_sim;         /* Complementarity with repeat sequence too hign.*/
    int ok;                 /* Number that were ok.                          */
} pair_stats;

typedef struct oligo_stats {
    int considered;         /* Total number of tested oligos of given type   */
    int ns;                 /* Number of oligos rejected because of Ns       */
    int target;             /* Overlapping targets.                          */
    int excluded;           /* Overlapping excluded regions.                 */
    int gc;                 /* Unappropriate GC content.                     */
    int gc_clamp;           /* Don't have required number of GCs at 3' end.  */
    int temp_min;           /* Melting temperature below t_min.              */
    int temp_max;           /* Melting temperature more than t_max.          */
    int compl_any;          /* Self-complementarity too high.                */
    int compl_end;          /* Self-complementarity at 3' end too high.      */
    int repeat;             /* Complementarity with repeat sequence too hign.*/
    int poly_x;             /* Long mononucleotide sequence inside.          */
    int seq_quality;        /* Low quality of bases included.                */
    int stability;          /* Stability of 5 3' bases too high.             */
    int ok;                 /* Number of acceptable oligos in this category. */
} oligo_stats;

/*
 * Arguments relating to a particular source sequence for which
 * we will pick primers.  Values for seq_args are _not_ retained
 * across different input records.
 */
typedef struct seqargs {
    pr_append_str error;    /* Error messages. */
    pr_append_str warning;  /* Warning messages. */
    int num_targets;        /* The number of targets. */
    interval_array_t tar;   /*
			     * The targets themselves; tar[i][0] is the start
			     * of the ith target, tar[i][1] its length.  These
			     * are presented as indexes within the sequence
			     * slot, but during the execution of choice() they
			     * are recalculated to be indexes within
			     * trimmed_seq.
			     */
    int num_excl;           /* The number of excluded regions.  */
    interval_array_t excl;  /* The same as for targets.
			     * These are presented as indexes within
			     * the sequence slot, but during the
			     * execution of choice() they are recalculated
			     * to be indexes within trimmed_seq.
			     */
    int num_internal_excl;  /* Number of excluded regions for internal oligo.*/
    interval_array_t excl_internal;
                            /* Similar to excl. */
    int incl_s;             /* The 0-based start of included region. */
    int incl_l;             /* 
			     * The length of the included region, which is
			     * also the length of the trimmed_seq field.
			     */
    int  *quality;          /* Vector of quality scores. */
    char *sequence;         /* The sequence itself. */
    char *sequence_name;    /* An identifier for the sequence. */
    char *sequence_file;    /* Another identifer for the sequence. */

    char *trimmed_seq;      /* The included region only. */
    char *left_input;       /* A left primer to check or design around. */
    char *right_input;      /* A right primer to check or design around. */
    char *internal_input;   /* An internal oligo to check or design around. */

    oligo_stats left_expl;  /* Left primers statistics. */
    oligo_stats right_expl; /* Right primers statistics. */
    oligo_stats intl_expl;  /* Internal oligos statistics. */
    pair_stats  pair_expl;  /* Pair statistics. */
} seq_args;

/* 
 * pr_program_name and pr_program_name_len must be set in the client's
 * main().
 */
extern const char *pr_program_name;

extern int pr_program_name_len;

/*
 * Return the ('\0'-terminated) sequence of oligo o in static storage that is
 * overwritten on each call.
 */
char *pr_oligo_sequence(const seq_args *, const primer_rec *o);

/*
 * Return the ('\0'-terminated) reverse-complemented sequence of oligo o in
 * static storage that is overwritten on each call.
 */
char *pr_oligo_rev_c_sequence(const seq_args *, const primer_rec *o);

void *pr_safe_malloc(size_t); /* A fail-stop wrapper for malloc. */

void *pr_safe_realloc(void *, size_t);

/* Append s to x. */
void pr_append(pr_append_str *, const char *);

/* Append s to x if x is empty.  Otherwise append concat("; ",s) to x. */
void pr_append_new_chunk(pr_append_str *, const char *s);
#endif

/* Set the argument to the empty string. */
void pr_set_empty(pr_append_str *);

/* Return true iff the argument is empty. */
int pr_is_empty(const pr_append_str *);

/* 
 * Gather all warnings and return a char containing them.  The return value
 * points to storage that the caller must free.  Return NULL if there are no
 * errors.
 */
char *pr_gather_warnings(const seq_args *, const primer_args *,
			 const primer_args *ma);

/* Choose primer pairs or triples */
void pr_choice(primer_args *, primer_args *, seq_args *, const dpal_args *,
	       const dpal_args *, const dpal_args *, pair_array_t *);

/* Print an "explanation string" for primer pair search. */
void pr_print_pair_explain(FILE *, const seq_args *);

void reverse_complement(const char *, char *);

