/* 
 * Copyright (c) 1996, 1997 Whitehead Institute for Biomedical Research. All
 * rights reserved.  Please see full software use agreement in primer3_main.c or
 * by executing primer3 with -h.
 */

#include <stdio.h>
#include <math.h>
#include <signal.h>
#include <unistd.h>
#include "primer3_release.h"
#include "format_output.h"
#include "dpal.h"
#include "oligotm.h"
#include "primer3.h"
#include "boulder_input.h"

/* Type definitions. */
typedef enum oligo_type { OT_LEFT = 0, OT_RIGHT = 1, OT_INTL = 2 }
oligo_type;

/* #define's */

/* 
 * Panic messages for when the program runs out of memory.  pr_program_name and
 * pr_program_name_len must be set at the beginning of main.
 */
#define OOM_MESSAGE      ": out of memory\n"
#define OOM_MESSAGE_LEN  16
#define OOM_ERROR \
    write(2, pr_program_name, pr_program_name_len), \
    write(2, OOM_MESSAGE, OOM_MESSAGE_LEN),   \
    exit(-2)

#ifndef MAX_PRIMER_LENGTH
#define MAX_PRIMER_LENGTH   35 /* 
				* Maximum primer length, governed largely by
				* the maximum size oligo for which one can
				* calculate melting temperature.
				*/
#endif

#define ALIGN_SCORE_UNDEF     SHRT_MIN

#define MACRO_CAT_2(A,B) A##B

#define INITIAL_LIST_LEN     2000 /* Initial size of oligo lists. */

#define PAIR_OK 1
#define PAIR_FAILED 0

#define DEFAULT_POSITION_PENALTIES(PA) \
    (PR_DEFAULT_INSIDE_PENALTY == pa->inside_penalty \
     && PR_DEFAULT_OUTSIDE_PENALTY == pa->outside_penalty)

/* Function declarations. */
static void   add_pair(const primer_pair *, pair_array_t *);
static short  align(const char *, const char*, const dpal_args *a);
static int    choose_pair(const primer_args *, const primer_args *,
			  seq_args *,
			  int, int, int, int,
			  const dpal_args*, const dpal_args*,
			  const dpal_args*,
			  pair_array_t *);
static int    check_intervals(const char *, const int,
			      interval_array_t, const int, seq_args *);
static int    choose_internal_oligo(const primer_rec *, const primer_rec *,
				    int, int *, seq_args *,
				    const primer_args *,
				    const dpal_args*,
				    const dpal_args *, const dpal_args *);
void          compute_position_penalty(const primer_args *, const seq_args *, 
				       primer_rec *, oligo_type);

static void   create_and_print_file(const seq_args *, int, const primer_rec[],
				    const oligo_type, const int, const int,
				    const char *);
static int    data_control(primer_args *, primer_args *, seq_args *);
static void   gc_and_n_content(const int, const int, const char *, primer_rec *);
static int    make_internal_oligos_list(const primer_args *,
					const primer_args *, seq_args *,
					int *, const dpal_args *,
					const dpal_args*,
					const dpal_args *);
static int    make_lists(const primer_args *, const primer_args *, 
			 seq_args *,
			 int *, int *, const dpal_args *, 
			 const dpal_args*, const dpal_args *);
static double obj_fn(const primer_args *, primer_pair *, int);
static int    oligo_overlaps_interval(const int, const int,
				      const interval_array_t, const int);
static int    oligo_pair_seen(const primer_pair *, const pair_array_t *);
static void   oligo_param(const primer_args *pa, const primer_args *ma,
			  primer_rec *, oligo_type,
			  const dpal_args*,
			  const dpal_args *, const dpal_args *,
			  seq_args *);
static int    pair_param(const primer_args *, seq_args *,
			 int, int, int, primer_pair *,
			 const dpal_args*,
			 const dpal_args *, const dpal_args*);
static int    pair_spans_target(const primer_pair *, const seq_args *);
static int    primer_pair_comp(const void *, const void*);
static int    primer_rec_comp(const void *, const void *);
static void   print_list(const seq_args *, const primer_args *,
			 const primer_args *, int, int, int);
static void   print_list_header(FILE *, oligo_type, int, int);
static void   print_oligo(FILE *, const seq_args *, int, const primer_rec *,
			  oligo_type, int, int);
static void   print_usage();
static void   boulder_print_pairs(const program_args *, const primer_args *,
       	       const primer_args *, const seq_args *, const pair_array_t *);
static void   print_explain(const oligo_stats *, oligo_type);
static void   print_all_explain(const primer_args *, const seq_args *);
static FILE   *safe_fopen(const char*, const char*);
static void   set_default_global_primer_args(primer_args *, primer_args *);
static void   set_dpal_args(dpal_args *);
static void   sig_handler(int);
static void   substr(const char *, int, int, char *);
static double tm(const primer_args *, const char*);
static double p_obj_fn(const primer_args *, const primer_args *, 
		  primer_rec *, int );
static void   oligo_compl(primer_rec *, const primer_args *, seq_args *,
			  oligo_type, const dpal_args *, 
			  const dpal_args*, const dpal_args *);
static int    pair_repeat_sim(primer_pair *, const primer_args *);

/* Global static variables. */
static const char * copyright[] = {
"",
"Copyright (c) 1996, 1997",
"        Whitehead Institute for Biomedical Research. All rights reserved.",
"",
"Redistribution and use in source and binary forms, with or without",
"modification, are permitted provided that the following conditions are met:",
"",
"1.      Redistributions must reproduce the above copyright notice, this",
"list of conditions and the following disclaimer in the  documentation",
"and/or other materials provided with the distribution.  Redistributions of",
"source code must also reproduce this information in the source code itself.",
"",
"2.      If the program is modified, redistributions must include a notice",
"(in the same places as above) indicating that the redistributed program is",
"not identical to the version distributed by Whitehead Institute.",
"",
"3.      All advertising materials mentioning features or use of this",
"software  must display the following acknowledgment:",
"        This product includes software developed by the",
"        Whitehead Institute for Biomedical Research.",
"",
"4.      The name of the Whitehead Institute may not be used to endorse or",
"promote products derived from this software without specific prior written",
"permission.",
"",
"We also request that use of this software be cited in publications as ",
"",
"Steve Rozen, Helen J. Skaletsky (1996)",
"   Primer3. Code available at",
"   http://www-genome.wi.mit.edu/genome_software/other/primer3.html",
"",
"THIS SOFTWARE IS PROVIDED BY THE WHITEHEAD INSTITUTE ``AS IS'' AND  ANY",
"EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  IMPLIED",
"WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  ARE",
"DISCLAIMED. IN NO EVENT SHALL THE WHITEHEAD INSTITUTE BE LIABLE  FOR ANY",
"DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL  DAMAGES",
"(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS  OR",
"SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)  HOWEVER",
"CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT",
"LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY",
"OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF",
"SUCH DAMAGE.",
NULL
};

#if 0
OLD CODE
This optimization made the program about 3sec faster
on the test suite.
static int *fn;   /* Number of Ns to the left of i. */
static int *fgc;  /* Number of GCs to the left of i. */
static int *fnn;  /* Number of non-Ns to the left of i including i */
#endif

/* Primer lists and their lengths. */
static primer_rec *f   = NULL;
static primer_rec *r   = NULL;
static primer_rec *mid = NULL;
static int f_len=0, r_len=0, mid_len=0;

/* Global variables. */
const char *pr_program_name;
int pr_program_name_len;

typedef struct oligo_array {
  int len;
  primer_rec *data;
} oligo_array;

#define FREE_STUFF { \
    free(sa); free(local_args); free(end_args);               \
    free(local_end_args);                                     \
    if (0 != best_pairs.storage_size) free(best_pairs.pairs); \
    free_seq_lib(&pa->repeat_lib); free(pa);                  \
    free_seq_lib(&ma->repeat_lib); free(ma);                  \
}

int
main(argc,argv)
    int argc;
    char *argv[]; 
{ 
    program_args prog_args;
    primer_args *pa;
    primer_args *ma;
    seq_args *sa;
    dpal_args *local_args;
    dpal_args *end_args;
    dpal_args *local_end_args;
    int input_found=0;
    pair_array_t best_pairs;

    pr_program_name = argv[0];
    pr_program_name_len = strlen(argv[0]);
    /* 
     * We set up some signal handlers in case someone starts up the program
     * from the command line, wonders why nothing is happening, and then kills
     * the program.
     */
    signal(SIGINT, sig_handler);
    signal(SIGTERM, sig_handler);

    /* 
     * We allocate the following structures on the heap rather than on the
     * stack in order to take advantage of testcenter memory access checking
     * during testing.
     */
    pa = pr_safe_malloc(sizeof(*pa));
    ma = pr_safe_malloc(sizeof(*ma));
    sa = pr_safe_malloc(sizeof(*sa));
    local_args = pr_safe_malloc(sizeof(*local_args));
    end_args = pr_safe_malloc(sizeof(*end_args));
    local_end_args = pr_safe_malloc(sizeof(*local_end_args));

    memset(pa, 0, sizeof(*pa));
    memset(ma, 0, sizeof(*ma));

    best_pairs.storage_size = best_pairs.num_pairs = 0;
    set_default_global_primer_args(pa, ma);

    memset(&prog_args, 0, sizeof(prog_args));
    while (--argc > 0) {
	argv++;
	if (!strcmp(*argv, "-format_output"))
	    prog_args.format_output = 1;
	else if (!strcmp(*argv, "-2x_compat"))
	    prog_args.twox_compat = 1;
	else if (!strcmp(*argv, "-strict_tags"))
	    prog_args.strict_tags = 1;
	else  {
	    print_usage();
	    FREE_STUFF;
	    exit(-1);
	}
    }

    set_dpal_args(local_args);
    local_args->flag = DPAL_LOCAL;
    set_dpal_args(end_args);
    end_args->flag = DPAL_GLOBAL_END;
    set_dpal_args(local_end_args);
    local_end_args->flag = DPAL_LOCAL_END;

    /* 
     * Read the data from input stream record by record and process it if
     * there are no errors.
     */
    while ((read_record(&prog_args, pa, ma, sa)) > 0) {
	input_found = 1;
	if (NULL == sa->error.data && NULL == pa->glob_err.data){
	    pr_choice(pa, ma, sa, local_args, end_args, local_end_args, 
							 &best_pairs);

        }
	if (NULL != pa->glob_err.data) 
	    pr_append_new_chunk(&sa->error, pa->glob_err.data);

	if (prog_args.format_output)
	    format_pairs(stdout, pa, ma, sa, &best_pairs);
        
	else
	    boulder_print_pairs(&prog_args, pa, ma, sa, &best_pairs);

	best_pairs.num_pairs = 0;

	if (NULL != sa->internal_input) free(sa->internal_input);
	if (NULL != sa->left_input) free(sa->left_input);
	if (NULL != sa->right_input) free(sa->right_input);
	if (NULL != sa->sequence) free(sa->sequence);
	if (NULL != sa->quality)  free(sa->quality);
	if (NULL != sa->trimmed_seq) free(sa->trimmed_seq);
	if (NULL != sa->sequence_name) free(sa->sequence_name);
	if (NULL != sa->error.data) free(sa->error.data);
	if (NULL != sa->warning.data) free(sa->warning.data);

	if (NULL != pa->glob_err.data) {
	    fprintf(stderr, "%s: %s\n", pr_program_name, pa->glob_err.data);
	    free(pa->glob_err.data);
	    FREE_STUFF;
	    exit(-4);
	}
    }
    FREE_STUFF;
    if (0 == input_found) {
	print_usage();
	exit(-3);
    }
    return 0;
}
#undef FREE_STUFF

/* Default parameter values.  */
#define OPT_SIZE            20
#define MIN_SIZE             18
#define MAX_SIZE             27

#define OPT_TM             60.0
#define MIN_TM             57.0
#define MAX_TM             63.0
#define MAX_DIFF_TM       100.0
#define MIN_GC             20.0
#define MAX_GC             80.0
#define SALT_CONC          50.0
#define DNA_CONC           50.0
#define NUM_NS_ACCEPTED       0
#define MAX_POLY_X            5
#define SELF_ANY            800
#define SELF_END            300
#define PAIR_COMPL_ANY      800
#define PAIR_COMPL_END      300
#define FILE_FLAG             0
/* 
 * COMPL_FLAG should be used only by program testers; it is 1 iff
 * complementarity is part of the objective function.
 */
#define COMPL_FLAG            0
#define EXPLAIN_FLAG          0
#define GC_CLAMP              0
#define LIBERAL_BASE          0
#define PICK_INTERNAL_OLIGO   0
#define INTERNAL_OLIGO_OPT_SIZE   20
#define INTERNAL_OLIGO_MIN_SIZE   18
#define INTERNAL_OLIGO_MAX_SIZE   27
#define INTERNAL_OLIGO_OPT_TM     60.0
#define INTERNAL_OLIGO_MIN_TM     57.0
#define INTERNAL_OLIGO_MAX_TM     63.0
#define INTERNAL_OLIGO_MIN_GC     20.0
#define INTERNAL_OLIGO_MAX_GC     80.0
#define INTERNAL_OLIGO_SALT_CONC         50.0
#define INTERNAL_OLIGO_DNA_CONC          50.0
#define INTERNAL_OLIGO_NUM_NS               0
#define INTERNAL_OLIGO_MAX_POLY_X           5 
#define INTERNAL_OLIGO_SELF_ANY          1200
#define INTERNAL_OLIGO_SELF_END          1200
#define INTERNAL_OLIGO_REPEAT_SIMILARITY 1200
#define REPEAT_SIMILARITY                1200
#define PAIR_REPEAT_SIMILARITY           2400
#define FIRST_BASE_INDEX            0
#define NUM_RETURN                  5
#define MIN_QUALITY                 0
#define QUALITY_RANGE_MIN           0
#define QUALITY_RANGE_MAX         100
#define DEFAULT_MAX_END_STABILITY     100.0

/* Assigns default values to primer picking parameters. */
static void
set_default_global_primer_args(a, b)
    primer_args *a, *b;
{
    a->primer_opt_size  = OPT_SIZE;
    a->primer_min_size  = MIN_SIZE;
    a->primer_max_size  = MAX_SIZE;
    a->opt_tm           = OPT_TM;
    a->min_tm           = MIN_TM;
    a->max_tm           = MAX_TM;
    a->max_diff_tm      = MAX_DIFF_TM;
    a->min_gc           = MIN_GC;
    a->max_gc           = MAX_GC;
    a->salt_conc        = SALT_CONC;
    a->dna_conc         = DNA_CONC;
    a->num_ns_accepted  = NUM_NS_ACCEPTED;
    a->self_any         = SELF_ANY;
    a->self_end         = SELF_END;
    a->pair_compl_any   = PAIR_COMPL_ANY;
    a->pair_compl_end   = PAIR_COMPL_END;
    a->file_flag        = FILE_FLAG;
    a->compl_flag       = COMPL_FLAG;
    a->explain_flag     = EXPLAIN_FLAG;
    a->gc_clamp         = GC_CLAMP;
    a->max_poly_x       = MAX_POLY_X;
    a->liberal_base        = LIBERAL_BASE;
    a->pick_internal_oligo = PICK_INTERNAL_OLIGO;
    a->first_base_index    = FIRST_BASE_INDEX;
    a->num_return            = NUM_RETURN;
    a->pr_min[0]           = 100;
    a->pr_max[0]           = 300;
    a->num_intervals      = 1;
    a->repeat_compl    = REPEAT_SIMILARITY;
    a->pair_repeat_compl  = PAIR_REPEAT_SIMILARITY;
    a->min_quality        = MIN_QUALITY;
    a->min_end_quality    = MIN_QUALITY;
    a->quality_range_min  = QUALITY_RANGE_MIN;
    a->quality_range_max  = QUALITY_RANGE_MAX;
    a->outside_penalty    = PR_DEFAULT_OUTSIDE_PENALTY;
    a->inside_penalty     = PR_DEFAULT_INSIDE_PENALTY;
    a->max_end_stability  = DEFAULT_MAX_END_STABILITY;

    b->primer_opt_size     = INTERNAL_OLIGO_OPT_SIZE;
    b->primer_min_size     = INTERNAL_OLIGO_MIN_SIZE;
    b->primer_max_size     = INTERNAL_OLIGO_MAX_SIZE;
    b->opt_tm          = INTERNAL_OLIGO_OPT_TM;
    b->min_tm          = INTERNAL_OLIGO_MIN_TM;
    b->max_tm          = INTERNAL_OLIGO_MAX_TM;
    b->min_gc          = INTERNAL_OLIGO_MIN_GC;
    b->max_gc          = INTERNAL_OLIGO_MAX_GC;
    b->max_poly_x      = INTERNAL_OLIGO_MAX_POLY_X;
    b->salt_conc       = INTERNAL_OLIGO_SALT_CONC;
    b->dna_conc        = INTERNAL_OLIGO_DNA_CONC;
    b->num_ns_accepted = INTERNAL_OLIGO_NUM_NS;
    b->self_any        = INTERNAL_OLIGO_SELF_ANY;
    b->self_end        = INTERNAL_OLIGO_SELF_END;
    b->repeat_compl    = INTERNAL_OLIGO_REPEAT_SIMILARITY;
    b->min_quality     = MIN_QUALITY;
    b->min_end_quality = MIN_QUALITY;
}
	
/* 
 * Finds up to pa->num_return primer pairs for the sequence seq with t targets.
 * Set sa->error and return 1 on error; otherwise return 0.
 */
void
pr_choice(pa, ma, sa, local_args, end_args, local_end_args, best_pairs)
    primer_args *pa, *ma;
    seq_args *sa;
    const dpal_args *local_args, *end_args, *local_end_args;
    pair_array_t *best_pairs;
{
    int i;       /* Loop index. */
    int n_f;     /* Number of acceptable left primers found. */
    int n_r;     /* Number of acceptable right primers found. */
    int n_m;     /* Number of acceptable internal oligos found. */
    int int_num; /* Product size range counter. */
    pair_array_t p;

    n_f = n_r = n_m = 0; 

    PR_ASSERT(NULL != pa);
    PR_ASSERT(NULL != ma);
    PR_ASSERT(NULL != sa);
    PR_ASSERT(NULL != local_args);
    PR_ASSERT(NULL != end_args);
    PR_ASSERT(NULL != local_end_args);

    if (data_control(pa, ma, sa) !=0 ) return;

    if (NULL == f) {
	f = pr_safe_malloc(sizeof(*f) * INITIAL_LIST_LEN);
	r = pr_safe_malloc(sizeof(*r) * INITIAL_LIST_LEN);
	f_len = r_len = INITIAL_LIST_LEN;
    }

    if (make_lists(pa, ma, sa, &n_f,&n_r,local_args, end_args, local_end_args)
								!=0 ) {
	return;
    }

    if (pa->pick_internal_oligo 
        && make_internal_oligos_list(pa, ma, sa, &n_m, local_args, 
				     end_args, local_end_args) != 0) {
	return;
    }

    /* Creates files with left, right, and internal oligos. */
    if (pa->file_flag) print_list(sa, pa, ma, n_f, n_r, n_m);

    /* We sort _after_ printing lists to maintain the order of test output. */
    qsort(&r[0], n_r, sizeof(*r), primer_rec_comp);
    qsort(&f[0], n_f, sizeof(*f), primer_rec_comp);

    p.storage_size = p.num_pairs = 0;

    /* Look for pa->num_return best primer pairs. */
    for(int_num=0; int_num < pa->num_intervals; int_num++) {
	if(choose_pair(pa, ma, sa, n_f, n_r, n_m, int_num,
		       local_args, end_args, local_end_args, &p)!=0)
           continue;

	for (i = 0; i < p.num_pairs && best_pairs->num_pairs < pa->num_return; i++)
	    if (!oligo_pair_seen(&p.pairs[i], best_pairs))
		add_pair(&p.pairs[i], best_pairs);

	if (pa->num_return == best_pairs->num_pairs) break;
	p.num_pairs = 0;
    }
    if (0 != p.storage_size) free(p.pairs);
}

/* Return 1 iff pair is already in the first num_pairs elements of retpair. */
static int
oligo_pair_seen(pair, retpair)
    const primer_pair *pair;
    const pair_array_t *retpair;
{
    const primer_pair *q = &retpair->pairs[0], 
		      *stop = &retpair->pairs[retpair->num_pairs];
    for (; q < stop; q++) {
	if (q->left->start == pair->left->start
	    && q->left->length == pair->left->length
	    && q->right->start == pair->right->start
	    && q->right->length == pair->right->length) return 1;
    }
    return 0;
}

/* Add 'pair' to 'retpair'. */
static void
add_pair(pair, retpair)
    const primer_pair *pair;
    pair_array_t *retpair;
{
    if (0 == retpair->storage_size) {
	retpair->storage_size = NUM_RETURN;
	retpair->pairs 
	    = pr_safe_malloc(retpair->storage_size * sizeof(*retpair->pairs));
    } else if (retpair->storage_size == retpair->num_pairs) {
	retpair->storage_size *= 2;
	retpair->pairs
	    = pr_safe_realloc(retpair->pairs,
			      retpair->storage_size * sizeof(*retpair->pairs));
    }
    retpair->pairs[retpair->num_pairs] = *pair;
    retpair->num_pairs++;
}

/* Make lists of acceptable left and right primers. */
static int
make_lists(pa, ma, sa, n_f, n_r, local_args, end_args, local_end_args)
    const primer_args *pa, *ma;
    seq_args *sa;
    int *n_f, *n_r;
    const dpal_args *local_args, *end_args, *local_end_args;
{
    int left, right;

    int i,j,n,k,pr_min;
    int tar_l, tar_r, f_b, r_b;
    char s[MAX_PRIMER_LENGTH+1],s1[MAX_PRIMER_LENGTH+1];
    primer_rec h;

    pr_min = INT_MAX;
    for(i=0;i<pa->num_intervals;i++)if(pa->pr_min[i]<pr_min)
	pr_min= pa->pr_min[i];

    PR_ASSERT(INT_MAX > (n=strlen(sa->trimmed_seq)));
    tar_r = 0;
    tar_l = n;
    for(i=0;i<sa->num_targets;i++) {
	if(sa->tar[i][0]>tar_r)tar_r = sa->tar[i][0];
	if(sa->tar[i][0]+sa->tar[i][1]-1<tar_l)tar_l=
	    sa->tar[i][0]+sa->tar[i][1]-1;
    }

    if (DEFAULT_POSITION_PENALTIES(pa)) {
      if (0 == tar_r) tar_r = n;
      if (tar_l == n) tar_l = 0;
    } else {
      tar_r = n;
      tar_l = 0;
    }

    if(tar_r - 1 < n - pr_min + pa->primer_max_size - 1)
      f_b=tar_r - 1;
    else
      f_b = n - pr_min + pa->primer_max_size-1;

    if (tar_l+1>pr_min - pa->primer_max_size)
      r_b = tar_l+1;
    else 
      r_b = pr_min - pa->primer_max_size;

    k=0;
    left=n; right=0;
    for (i = f_b; i >= pa->primer_min_size - 1; i--) {
	s[0]='\0';
	for (j = pa->primer_min_size; j <= pa->primer_max_size; j++) {
	    if (i-j > n-pr_min-1) continue;
	    if (i-j+1>=0) {
		if (k >= f_len) {
		    f_len += (f_len >> 1);
		    f = pr_safe_realloc(f, f_len * sizeof(*f));
		}
		f[k].start=i-j+1;
		f[k].length=j;
		substr(sa->trimmed_seq,f[k].start,f[k].length,s);

		if (sa->left_input && strcmp(sa->left_input, s))
		  continue;

		h.start = f[k].start; h.length = f[k].length;
		h.repeat_sim.score = NULL;
		oligo_param(pa, ma, &h, OT_LEFT, local_args, end_args, 
				     local_end_args, sa);
		if (0 == h.ok) h.quality = p_obj_fn(pa, ma, &h, 0);
		if (pa->explain_flag) {
		    sa->left_expl.considered++;
		    switch (h.ok)
		    {
		    case 1 : sa->left_expl.ns++;          break;
		    case 2 : sa->left_expl.target++;      break;
		    case 3 : sa->left_expl.gc++;          break;
		    case 4 : sa->left_expl.temp_min++;    break;
		    case 5 : sa->left_expl.temp_max++;    break;
		    case 8 : sa->left_expl.excluded++;    break;
		    case 9 : sa->left_expl.gc_clamp++;    break;
		    case 10: sa->left_expl.stability++;   break;
		    case 11: sa->left_expl.poly_x++;      break;
		    case 12: sa->left_expl.seq_quality++; break;
		    }
               }
		if (h.ok==1||h.ok==2||h.ok==6|| /* h.ok==7|| */
	                    h.ok==11||h.ok==10||h.ok==8||h.ok==9
                            || h.ok == 10) break;
		else{
		    if (h.ok==0) {
			f[k] = h;
			if(f[k].start<left) left=f[k].start;
			k++;
		    }
		}
	    }
	    else break;
	}
    }
    *n_f = k;
    k=0;
    for(i=r_b; i<=n-pa->primer_min_size; i++) {
	s[0]='\0';
	for(j = pa->primer_min_size; j <= pa->primer_max_size; j++) {
	    if(i+j<pr_min) continue;
	    if(i+j-1<n){
		if (k >= r_len) {
		    r_len += (r_len >> 1);
		    r = pr_safe_realloc(r, r_len * sizeof(*r));
		}
		r[k].start=i+j-1;
		r[k].length=j;
		substr(sa->trimmed_seq,i,j,s);
		reverse_complement(s,s1);

		if (sa->right_input && strcmp(sa->right_input, s1))
		  continue;

		h.start = r[k].start; h.length = r[k].length;
		h.repeat_sim.score = NULL;
		oligo_param(pa, ma, &h, OT_RIGHT, local_args, end_args, 
					  local_end_args, sa);
		if (0 == h.ok) h.quality = p_obj_fn(pa, ma, &h, 0);
		if (pa->explain_flag){
		    sa->right_expl.considered++;
		    switch (h.ok)
		    {
		    case 1 : sa->right_expl.ns++;           break;
		    case 2 : sa->right_expl.target++;       break;
		    case 3 : sa->right_expl.gc++;           break;
		    case 4 : sa->right_expl.temp_min++;     break;
		    case 5 : sa->right_expl.temp_max++;     break;
		    case 8 : sa->right_expl.excluded++;     break;
		    case 9 : sa->right_expl.gc_clamp++;     break;
		    case 10: sa->right_expl.stability++;    break;
		    case 11: sa->right_expl.poly_x++;       break;
		    case 12: sa->right_expl.seq_quality++;  break;
		    }
                }

		if (h.ok==1||h.ok==2||h.ok==6|| /* h.ok==7|| */
		    h.ok==10||h.ok==11||h.ok==8||h.ok==9
		    ||h.ok == 10) break;
		else{
		    if(h.ok==0){
			r[k] = h;
			if(r[k].start>right) right=r[k].start;
			k++;
		    }
		}
	    }
	    else break;
	}
    }
    *n_r=k;

    /* 
     * Return 1 if one of lists is empty or if leftmost left primer and
     * rightmost right primer do not provide sufficient product size.
     */
    sa->left_expl.ok = *n_f;
    sa->right_expl.ok = *n_r;
    if (0 == *n_f || 0 == *n_r) return 1;
    else if (right - left < pr_min - 1) {
	sa->pair_expl.product    = 1;
	sa->pair_expl.considered = 1;
	return 1;
    } else return 0;
}

/* 
 * Make complete list of acceptable internal oligos in mid.  Place the number
 * of valid elements in mid in *n_m.  Return 1 if there are no acceptable
 * internal oligos; otherwise return 0.
 */
static int
make_internal_oligos_list(pa, ma, sa, n_m, local_args, end_args, local_end_args)
    const primer_args *pa, *ma;
    seq_args *sa;
    int *n_m;
    const dpal_args *local_args, *end_args, *local_end_args;
{
    int i,j,n,k;

    char s[MAX_PRIMER_LENGTH+1];
    primer_rec h;

    if (NULL == mid) {
	mid_len = INITIAL_LIST_LEN;
	mid = pr_safe_malloc(sizeof(*mid) * mid_len);
    }

    n = strlen(sa->trimmed_seq);
    k = 0;
    for(i = n - 1; i >= ma->primer_min_size-1; i--) {
	s[0] = '\0';
	for(j = ma->primer_min_size; j <=ma->primer_max_size; j++) {
	    if(i-j < -1) break;
	    if (k >= mid_len) {
		mid_len += (mid_len >> 1);
		mid = pr_safe_realloc(mid, mid_len * sizeof(*mid));
	    }
	    mid[k].start = i - j +1;
	    mid[k].length = j;
	    substr(sa->trimmed_seq, mid[k].start, mid[k].length,s);

	    if (sa->internal_input && strcmp(sa->internal_input, s))
	      continue;

	    h.start = mid[k].start; h.length = mid[k].length;
	    h.repeat_sim.score = NULL;
	    oligo_param(pa, ma, &h, OT_INTL, local_args, end_args, 
					      local_end_args, sa);
	    if (0 == h.ok) h.quality = p_obj_fn(pa, ma, &h, 2);
	    if (pa->explain_flag){
		sa->intl_expl.considered++;
		switch (h.ok)
		{
		case 1 : sa->intl_expl.ns++;             break;
		case 2 : sa->intl_expl.target++;         break;
		case 3 : sa->intl_expl.gc++;             break;
		case 4 : sa->intl_expl.temp_min++;       break;
		case 5 : sa->intl_expl.temp_max++;       break;
		case 8 : sa->intl_expl.excluded++;       break;
		case 9 : sa->intl_expl.gc_clamp++;       break;
		case 11: sa->intl_expl.poly_x++;         break;
		case 12: sa->intl_expl.seq_quality++;    break;
		}
            }
	    if (h.ok==1||h.ok==2||h.ok==6 /* ||h.ok==7 */ ||h.ok==8
		       ||h.ok==11||h.ok==10) break;
	    else {
		if(h.ok==0) {
		    mid[k] = h;
		    k++;
		}
           }
       }
    }
    *n_m = k;
    sa->intl_expl.ok = *n_m;
    if(*n_m==0)return 1;
    else return 0;
}

/* Takes substring of seq starting from n with length m and puts it to s.    */
static void
substr(const char *seq, int n, int m, char *s)
{
	int i;
	for(i=n;i<n+m;i++)s[i-n]=seq[i];
	s[m]='\0';
}

/* Reverse and complement the sequence seq and put the result in s. */ 
void
reverse_complement(const char *seq, char *s)
{
    const char *p = seq;
    char *q = s;

    while (*p != '\0') p++;
    p--;
    while (p >= seq) {
	switch (*p)
	{
	case 'A': *q='T'; break;
	case 'C': *q='G'; break;
	case 'G': *q='C'; break;
	case 'T': *q='A'; break;
	case 'U': *q='A'; break;

	case 'B': *q='V'; break;
	case 'D': *q='H'; break;
        case 'H': *q='D'; break;
        case 'V': *q='B'; break;
        case 'R': *q='Y'; break;
        case 'Y': *q='R'; break;
	case 'K': *q='M'; break;
        case 'M': *q='K'; break;
        case 'S': *q='S'; break;
        case 'W': *q='W'; break;

	case 'N': *q='N'; break;

	case 'a': *q='t'; break;
	case 'c': *q='g'; break;
	case 'g': *q='c'; break;
	case 't': *q='a'; break;
	case 'u': *q='a'; break;

	case 'b': *q='v'; break;
	case 'd': *q='h'; break;
        case 'h': *q='d'; break;
        case 'v': *q='b'; break;
        case 'r': *q='y'; break;
        case 'y': *q='r'; break;
	case 'k': *q='m'; break;
        case 'm': *q='k'; break;
        case 's': *q='s'; break;
        case 'w': *q='w'; break;

	case 'n': *q='n'; break;
	}
	p--; q++;
    }
    *q = '\0';
}

/*
 * Define number of Ns, GC content, intersections with targets, melting
 * temperature and self complementarity of one oligo.
 *
 * Checks if given primer satisfies the requirements. Assigns ok 0 if primer is
 * acceptable, 1 if number of Ns is too large, 2 if primer intersects target,
 * 3 if GC content is out of range, 4 if temperature is too low, 5 if too high,
 * 6 or 7 if self complementarity is too large. 
 * Assigns ok = 8 if oligo is in an excluded region.
 *         ok = 9 if gc clamp is condition is violated.
 *         ok = 10 if 3' max stability condition is violated.
 *
 * If ok=0, calculate part of objective function due to this primer.
 */
static void
oligo_param(pa, ma, h, l, local_args, end_args, local_end_args, sa)
    const primer_args *pa, *ma;
    primer_rec *h;
    oligo_type l;
    const dpal_args *local_args, *end_args, *local_end_args;
    seq_args *sa;
{
    int i,j,k, m, qu, qu_end;
    int poly_x, max_poly_x;
    const char *seq = sa->trimmed_seq;
    char s1[MAX_PRIMER_LENGTH+1], s1_rev[MAX_PRIMER_LENGTH+1];

    h->ok = -1;
    h->target = h->gc_content = h->num_ns=h->excl=0;

    PR_ASSERT(OT_LEFT == l || OT_RIGHT == l || OT_INTL == l);
    
    if (OT_LEFT == l || OT_INTL == l) {j = h->start; k=j+h->length-1;}
    else {j = h->start-h->length+1; k=h->start;}

    PR_ASSERT(k >= 0);
    PR_ASSERT(k < TRIMMED_SEQ_LEN(sa));

    gc_and_n_content(j, k-j+1, sa->trimmed_seq, h);

    if (((OT_LEFT == l || OT_RIGHT == l) && h->num_ns > pa->num_ns_accepted)
	|| (OT_INTL == l && h->num_ns > ma->num_ns_accepted)) {
	h->ok = 1;
	return;
    }

    /* Upstream error checking has ensured that we use non-default position
       penalties only when there is 0 or 1 target. */
    PR_ASSERT(sa->num_targets <= 1 || DEFAULT_POSITION_PENALTIES(pa));
    if (l < 2 
	&& DEFAULT_POSITION_PENALTIES(pa)
	&& oligo_overlaps_interval(j, k-j+1, sa->tar, sa->num_targets))
      h->target = 1;
    else if (l < 2 && !DEFAULT_POSITION_PENALTIES(pa)
	     && 1 == sa->num_targets) {
      compute_position_penalty(pa, sa, h, l);
      if (h->position_penalty_infinite) h->target = 1;
    } else {
      h->position_penalty = 0.0;
      h->position_penalty_infinite = '\0';
    }

    if (l < 2 && oligo_overlaps_interval(j, k-j+1, sa->excl, sa->num_excl))
	h->excl = 1;

    if (l == 2 && oligo_overlaps_interval(j, k-j+1, sa->excl_internal,
					  sa->num_internal_excl))
	h->excl = 1;

    if(l < 2 && h->target==1) {
	h->ok=2;
	return;
    }

    if(h->excl==1){
	h->ok=8;
	return;
    }

    if((l<2 && (h->gc_content< pa->min_gc || h->gc_content > pa->max_gc))||
      (l==2 && (h->gc_content< ma->min_gc || h->gc_content > ma->max_gc))){
	h->ok = 3;
	return;
    }
    if(pa->gc_clamp != 0){
       if(OT_LEFT == l){
	   for(i=k-pa->gc_clamp+1; i<= k; i++)if(seq[i] !='G'&&seq[i] !='C'){
	       h->ok = 9;
	       return;
           }
       }
       if(OT_RIGHT == l){
	   for(i=j; i<j+pa->gc_clamp; i++)if(seq[i] != 'G' && seq[i] != 'C'){
	       h->ok = 9;
	       return;
           }
       }
    }
            
    if((pa->min_quality != 0 && (OT_LEFT == l || OT_RIGHT == l)) ||
       (ma->min_quality != 0 && OT_INTL == l)) {
       if(OT_LEFT == l || OT_RIGHT == l){
	  qu = pa->min_quality;
	  qu_end = pa->min_end_quality;
       }
       else {qu = qu_end = ma->min_quality;}
       h->seq_quality = pa->quality_range_max;
       if(OT_LEFT == l || OT_INTL == l){
	  for(i = j; i<=k-5; i++) {
	     if((m = sa->quality[i + sa->incl_s]) < qu){
		h->ok = 12;
		return;
             }
	     else if (m < h->seq_quality) h->seq_quality = m;
          }
	  for(i = k-4; i <= k; i++) {
	     if(i < j) continue;
	     if((m = sa->quality[i + sa->incl_s]) < qu_end){
		h->ok = 12;
		return;
             }
	     else if (m < h->seq_quality) h->seq_quality = m;
          }
       }
       else {
	  for(i = j; i < j+5; i++) {
	     if(i > k) break;
	     if((m =sa->quality[i + sa->incl_s]) < qu_end){
		h->ok = 12;
		return;
             }
	     else if (m < h->seq_quality) h->seq_quality = m;
          }
	  for(i = j+5; i <=k; i++) {
	     if((m = sa->quality[i + sa->incl_s]) < qu){
		h->ok = 12;
		return;
             }
	     else if (m < h->seq_quality) h->seq_quality = m;
          }
       }
    }

    if(OT_LEFT == l || OT_RIGHT == l) max_poly_x = pa->max_poly_x;     
    else max_poly_x = ma->max_poly_x;
    if(max_poly_x > 0){
          poly_x = 1;
	  for(i=j+1;i<=k;i++){
                if(seq[i] == seq[i-1]||seq[i] == 'N'){
		       poly_x++;
		       if(poly_x > max_poly_x){
			     h->ok = 11;
			     return;
                       }
                }
		else poly_x = 1;
          }
     }

    substr(seq,j,k-j+1,s1);
    if(OT_LEFT == l || OT_RIGHT == l) h->temp = tm(pa, s1);
    else h->temp = tm(ma, s1);
    if ((l<2&&h->temp<pa->min_tm)||(l==2&&h->temp<ma->min_tm)) {
	h->ok=4;
	return;
    }
    if (((OT_LEFT == l || OT_RIGHT == l) && h->temp>pa->max_tm) 
	|| (OT_INTL == l && h->temp>ma->max_tm)) {
	h->ok=5;
	return;
    }
    if (OT_LEFT == l) {
      if ((h->end_stability = end_oligodg(s1, 5))
	  > pa->max_end_stability) {
	h->ok = 10;
	return;
      }
    } else if (OT_RIGHT == l) {
      reverse_complement(s1, s1_rev);
      if ((h->end_stability = end_oligodg(s1_rev, 5))
	  > pa->max_end_stability) {
	  h->ok = 10;
	  return;
      }
    }

    if(pa->compl_flag == 1){
       if(OT_LEFT == l || OT_RIGHT == l)
		     oligo_compl(h, pa, sa, l, local_args, end_args, 
						    local_end_args);
       else          oligo_compl(h, ma, sa, l, local_args, end_args,
						    local_end_args);
       if(h->ok != 0) return;
    }
    else h->self_any = h->self_end  = ALIGN_SCORE_UNDEF; 
    h->ok = 0;
}

/* 
 * Set h->gc_content to the GC % of the 'len' bases starting at 'start' in
 * 'sequence' (excluding 'N's).  Set h->num_ns to the number of 'N's in that
 * subsequence.
 */
static void
gc_and_n_content(start, len, sequence, h)
    const int start, len;
    const char *sequence;
    primer_rec *h;
{
    const char* p = &sequence[start];
    const char* stop = p + len;
    int num_gc = 0, num_gcat = 0, num_n = 0;
    while (p < stop) {
	if ('N' == *p) 
	    num_n++;
	else {
	    num_gcat++;
	    if ('C' == *p || 'G' == *p) num_gc++;
	}
	p++;
    }
    h->num_ns = num_n;
    if (0 == num_gcat) h->gc_content= 0.0;
    else h->gc_content = 100.0 * ((double)num_gc)/num_gcat;
}

static int
oligo_overlaps_interval(start, len, intervals, num_intervals)
    const int start, len;
    const interval_array_t intervals;
    const int num_intervals;
{
    int i;
    int last = start + len - 1;
    for (i = 0; i < num_intervals; i++)
	if (!(last < intervals[i][0] 
	      || start > (intervals[i][0] + intervals[i][1] - 1)))
	    return 1;
    return 0;
}

/* Calculate the part of the objective function due to one primer. */
static double
p_obj_fn(pa, ma, h, j)
    const primer_args *pa, *ma;
    primer_rec *h;
    int  j;
{
  if (j >= 2)
    return fabs(h->temp - ma->opt_tm) + abs(h->length - ma->primer_opt_size);
  else {
    PR_ASSERT(!h->position_penalty_infinite);
    return h->position_penalty + fabs(h->temp - pa->opt_tm)
	    + abs(h->length - pa->primer_opt_size);
  }
}

static double
tm(pa, s)
    const primer_args *pa;
    const char *s;
{
  return oligotm(s, pa->dna_conc, pa->salt_conc);
}

static void
print_list(sa, pa, ma, n_f, n_r, n_m)
    const seq_args *sa;
    const primer_args *pa, *ma;
    int n_f, n_r, n_m;
{
    int first_base_index = pa->first_base_index;
    create_and_print_file(sa, n_f, f, OT_LEFT, first_base_index, 
			  NULL != pa->repeat_lib.repeat_file, ".for");
    create_and_print_file(sa, n_r, r, OT_RIGHT, first_base_index,
			  NULL != pa->repeat_lib.repeat_file, ".rev");
    if (pa->pick_internal_oligo)
      create_and_print_file(sa, n_m, mid, OT_INTL,
			    first_base_index,
			    NULL != ma->repeat_lib.repeat_file, ".int");
}

static void
create_and_print_file(sa, n, oligo_arr, o_type,
		      first_base_index, print_lib_sim, ext)
    const seq_args *sa;
    int n;
    const primer_rec oligo_arr[];
    const oligo_type o_type;
    const int first_base_index, print_lib_sim;
    const char *ext;
{
    int i;
    char *file = pr_safe_malloc(strlen(sa->sequence_name) + strlen(ext) + 1);
    FILE *fh;

    strcpy(file, sa->sequence_name);
    strcat(file,ext);
    fh = safe_fopen(file,"w");
    print_list_header(fh, o_type, first_base_index, print_lib_sim);
    for(i=0; i<n; i++) 
	print_oligo(fh, sa, i, &oligo_arr[i], o_type,
		    first_base_index, print_lib_sim);
    fclose(fh);
    free(file);
}

static void
print_list_header(fh, type, first_base_index, print_lib_sim)
    FILE *fh;
    oligo_type type;
    int first_base_index, print_lib_sim;
{
    fprintf(fh, "ACCEPTABLE %s\n",
            OT_LEFT == type ? "LEFT PRIMERS"
            : OT_RIGHT == type ? "RIGHT PRIMERS" : "INTERNAL OLIGOS");
    fprintf(fh, "                               %4d-based     ", 
	    first_base_index);
    if (print_lib_sim)
	fprintf(fh, "#               self  self   lib  qual-\n");
    else
	fprintf(fh, "#               self  self  qual-\n");
    fprintf(fh, "   # sequence                       start ln  "); 
    if (print_lib_sim)
	fprintf(fh, "N   GC%%     Tm   any   end   sim   lity\n");
    else 
	fprintf(fh, "N   GC%%     Tm   any   end   lity\n");
}

static void
print_oligo(fh, sa, index, h, type, first_base_index, print_lib_sim)
    FILE *fh;
    const seq_args *sa;
    int index;
    const primer_rec *h;
    oligo_type type;
    int first_base_index, print_lib_sim;
{
    char *p = (OT_RIGHT != type) 
	? pr_oligo_sequence(sa, h) : pr_oligo_rev_c_sequence(sa, h);
    if (print_lib_sim)
	fprintf(fh,
		"%4d %-30s %5d %2d %2d %5.2f %5.3f %5.2f %5.2f %5.2f %6.3f\n",
		index, p, h->start+sa->incl_s + first_base_index, 
		h->length,
		h->num_ns, h->gc_content, h->temp,
		h->self_any / PR_ALIGN_SCORE_PRECISION,
		h->self_end / PR_ALIGN_SCORE_PRECISION,
		h->repeat_sim.score[h->repeat_sim.max] /PR_ALIGN_SCORE_PRECISION,
		h->quality);
    else
	fprintf(fh,
		"%4d %-30s %5d %2d %2d %5.2f %5.3f %5.2f %5.2f %6.3f\n",
		index, p, h->start+sa->incl_s + first_base_index, 
		h->length,
		h->num_ns, h->gc_content, h->temp,
		h->self_any / PR_ALIGN_SCORE_PRECISION,
		h->self_end / PR_ALIGN_SCORE_PRECISION,
		h->quality);
}

static int
choose_pair(pa, ma, sa, n_f, n_r, n_m, int_num, local_args, end_args, 
					      local_end_args, p)
    const primer_args *pa, *ma;
    seq_args *sa;
    int n_f;
    int n_r;
    int n_m;
    int int_num;
    const dpal_args *local_args, *end_args, *local_end_args;
    pair_array_t *p;
{
    int i,j;
    int k; /* 
	    * The number of acceptable primer pairs saved in p.
	    * (k <= pa->num_return.)
	    */
    int i0, n_last, n_int;
    primer_pair worst_pair; /* The worst pair among those being "remembered". */
    int i_worst;             /* The index within p of worst_pair. */

    primer_pair h;

    k=0; 

    i_worst = 0;
    n_last = n_f;
    for(i=0;i<n_r;i++) {
	/* 
	 * Make a quick cut based on the the quality of the best left
	 * primer.
	 *
	 * worst_pair must be defined if k >= pa->num_return.
	 */
	if(r[i].ok != 0) continue;
	if (k >= pa->num_return 
	    && r[i].quality+f[0].quality > worst_pair.pair_quality)
	    break;

	for(j=0;j<n_last;j++) {
	    /* 
	     * Invariant: if 2 pairs in p have the same pair_quality, then the
	     * pair discovered second will have a higher index within p.
	     *
	     * This invariant is needed to produce consistent results when 2
	     * pairs have the same pair_quality.
	     */

	    if(r[i].ok != 0) break;
	    if(f[j].ok != 0) continue;
	    if(k>= pa->num_return 
	       && f[j].quality+r[i].quality > worst_pair.pair_quality) {
		/* worst_pair must be defined if k >= pa->num_return. */
		n_last=j;
		break;
	    }

	    if (PAIR_OK ==
		pair_param(pa, sa, j, i, int_num, &h, local_args, end_args,
							  local_end_args)) {
		h.pair_quality = obj_fn(pa, &h, int_num);

		if (pa->pick_internal_oligo
		    && choose_internal_oligo(h.left, h.right,
			     n_m, &n_int, sa, ma, local_args, end_args,
						     local_end_args)!=0){
		    sa->pair_expl.internal++;
		    continue;
                }
		sa->pair_expl.ok++;
		if (k < pa->num_return) {
		    if (pa->pick_internal_oligo) h.intl = &mid[n_int];
		    add_pair(&h, p);
		    /* p->pairs[k] = h; */
		    if (k == 0 || primer_pair_comp(&h, &worst_pair) > 0){
			worst_pair = h;
                        i_worst = k;
                    }
		    k++;
		} else if (primer_pair_comp(&h, &worst_pair) < 0) {
		    /* 
		     * There are already pa->num_return results, and vl is better than
		     * the pa->num_return the quality found so far.
		     */
		    if (pa->pick_internal_oligo) h.intl = &mid[n_int];
		    p->pairs[i_worst] = h;
		    worst_pair = h; /* h is a lower bound on the worst pair. */
		    for (i0 = 0; i0<pa->num_return; i0++)
			if(primer_pair_comp(&p->pairs[i0], &worst_pair) > 0) {
			    i_worst = i0;
			    worst_pair = p->pairs[i0];

			}
		}
	    }
	}	
    }
    if(k!=0) qsort(p->pairs, k, sizeof(primer_pair), primer_pair_comp);
    p->num_pairs = k;
    if (k==0) return 1;
    else return 0;
}

/* Choose best internal oligo for given pair of left and right primers. */
static int
choose_internal_oligo(left, right, n_m, nm, sa, ma, local_args, 
					end_args,  local_end_args)
     const primer_rec *left, *right;
     int n_m;
     int *nm;
     seq_args *sa;
     const primer_args *ma;
     const dpal_args *local_args, *end_args, *local_end_args;
{
   int i,k;
   double min;

   min = 1000000.;
   i = -1;
   for(k=0; k < n_m; k++){
       if(mid[k].start > left->start + left->length-1&&
	     mid[k].start+mid[k].length-1 < right->start-right->length+1&&
		   mid[k].quality < min && mid[k].ok == 0) {
		       if(mid[k].self_any == ALIGN_SCORE_UNDEF){
			   oligo_compl(&mid[k], ma, sa, OT_INTL, local_args, 
					    end_args, local_end_args);
			   if(mid[k].ok != 0) continue;
                       }
		       min = mid[k].quality;
		       i=k;
       }    
    }
    *nm = i;
    if(*nm < 0) return 1;
    return 0;
}

/* Compare function for sorting primer records. */
static int
primer_rec_comp(x1, x2)
    const void *x1, *x2;
{
    const primer_rec *a1 = x1, *a2 = x2;

    if(a1->quality < a2->quality) return -1;
    if (a1->quality > a2->quality) return 1;

    /* 
     * Even assuming that self-complementarity is not part of the objective
     * function, we still perfer primers with less self-complementarity.  1.1
     * is an arbitrary factor to give more weight to 3'
     * self-complementarity.
    w1 = a1->self_any + 1.1 * a1->self_end;
    w2 = a2->self_any + 1.1 * a2->self_end;
    if (w1 < w2) return -1;
    if (w1 > w2) return 1;
    */

    /* 
     * We want primer_rec_comp to always return a non-0 result, because
     * different implementations of qsort differ in how they treat "equal"
     * elements, making it difficult to compare test output on different
     * systems.
     */
     if(a1->start > a2->start) return -1;
     if(a1->start < a2->start) return 1;
     if(a1->length < a2->length) return -1;
     return 1;
}

/* Compare function for sorting primer records. */
static int
primer_pair_comp(x1, x2)
    const void *x1, *x2;
{
    const primer_pair *a1 = x1, *a2 = x2;
    int y1, y2;

    if(a1->pair_quality < a2->pair_quality) return -1;
    if (a1->pair_quality > a2->pair_quality) return 1;

    if (a1->compl_measure < a2->compl_measure) return -1;
    if (a1->compl_measure > a2->compl_measure) return 1;

    /* 
     * We want primer_pair_comp to always return the pair with the rightmost
     * left primer, all other measures being equal, becuase different
     * implementations of qsort differ in how they treat "equal" elements,
     * making it difficult to compare test output on different systems.
     */

    y1 = a1->left->start;
    y2 = a2->left->start;
    if (y1 > y2) return -1;  /* prefer left primers to the right. */
    if (y1 < y2) return 1;
    
    y1 = a1->right->start;
    y2 = a2->right->start;
    if (y1 < y2) return -1; /* prefer right primers to the left. */
    if (y1 > y2) return 1; 

    y1 = a1->left->length;
    y2 = a2->left->length;
    if (y1 < y2) return -1;  /* prefer shorter primers. */
    if (y1 > y2) return 1;
    
    y1 = a1->right->length;
    y2 = a2->right->length;
    if (y1 < y2) return -1; /* prefer shorter primers. */
    if (y1 > y2) return 1; 
    
    return 0;
}

/* 
 * Defines parameter values for given primer pair. Returns PAIR_OK if the pair is
 * acceptable; PAIR_FAILED otherwise.
 */
static int
pair_param(pa, sa, m, n, int_num, h, local_args, end_args, local_end_args)
    const primer_args *pa;
    seq_args *sa;
    int m, n, int_num;
    primer_pair *h;
    const dpal_args *local_args, *end_args, *local_end_args;
{
    char s1[MAX_PRIMER_LENGTH+1], s2[MAX_PRIMER_LENGTH+1], 
    s1_rev[MAX_PRIMER_LENGTH+1], s2_rev[MAX_PRIMER_LENGTH+1];
    short compl_end;

    h->left = &f[m];
    h->right = &r[n];
    h->product_size = r[n].start - f[m].start+1;
    h->target = 0;
    h->compl_any = h->compl_end = 0;
    sa->pair_expl.considered++;

    if(h->product_size < pa->pr_min[int_num] || 
		h->product_size > pa->pr_max[int_num]) {
	sa->pair_expl.product++;
	h->product_size = -1;
	return PAIR_FAILED;
    }

    if (sa->num_targets > 0) {
	if (pair_spans_target(h, sa))
	    h->target = 1;
	else {
	    h->target = -1;
	    sa->pair_expl.target++;
	    return PAIR_FAILED;
	}
    }

    h->diff_tm = fabs(f[m].temp - r[n].temp);
    if (h->diff_tm > pa->max_diff_tm) {
	sa->pair_expl.temp_diff++;
	return PAIR_FAILED;
    }

    substr(sa->trimmed_seq,f[m].start,f[m].length,s1);
    substr(sa->trimmed_seq,r[n].start-r[n].length+1,r[n].length,s2);
    if(f[m].self_any == ALIGN_SCORE_UNDEF){
       oligo_compl(&f[m], pa, sa, OT_LEFT, local_args, end_args, local_end_args);
       if(f[m].ok != 0) {
	   sa->pair_expl.considered--;
	   return PAIR_FAILED;
       }
    }
    if(r[n].self_any == ALIGN_SCORE_UNDEF){
       oligo_compl(&r[n], pa, sa, OT_RIGHT, local_args, end_args, local_end_args);
       if(r[n].ok != 0) {
	  sa->pair_expl.considered--;
	  return PAIR_FAILED;
       }
    }
	
    /* 
     * Similarity between s1 and s2 is equivalent to complementarity between
     * s2's complement and s1.  (Both s1 and s2 are taken from the same strand.)
     */
    h->compl_any = align(s1,s2,local_args);
    if (h->compl_any > pa->self_any) {
	sa->pair_expl.compl_any++;
	return PAIR_FAILED;
    }

    if ((h->compl_end = align(s1,s2,end_args)) > pa->self_end) {
	    sa->pair_expl.compl_end++;
	    return PAIR_FAILED;
    }
    /*
     * It is conceivable (though very unlikely in practice) that
     * align(s2_rev, s1_rev, end_args) > align(s1,s2,end_args).
     */
    reverse_complement(s1, s1_rev);
    reverse_complement(s2, s2_rev);
    if((compl_end = align(s2_rev, s1_rev, end_args)) > h->compl_end)  {
	if (compl_end > pa->self_end) {
	    sa->pair_expl.compl_end++;
	    return PAIR_FAILED;
	}
	h->compl_end = compl_end;
    }
    h->compl_measure = 	
	(h->right->self_end  + h->left->self_end + h->compl_end) * 1.1
	    + h->right->self_any + h->left->self_any + h->compl_any;
    if((h->repeat_sim = pair_repeat_sim(h, pa)) > pa->pair_repeat_compl){
	 sa->pair_expl.repeat_sim++;
	 return PAIR_FAILED;
    }
    return PAIR_OK;
}

/* compute_position_penalty is experimental code. */
void
compute_position_penalty(pa, sa, h, o_type)
  const primer_args *pa;
  const seq_args *sa;
  primer_rec *h;
  oligo_type o_type;
{
  int three_prime_base;
  int inside_flag = 0;
  int target_begin, target_end;

  PR_ASSERT(OT_LEFT == o_type || OT_RIGHT == o_type);
  PR_ASSERT(1 == sa->num_targets);
  target_begin = sa->tar[0][0];
  target_end = target_begin + sa->tar[0][1] - 1;

  three_prime_base = OT_LEFT == o_type
    ? h->start + h->length - 1 : h->start - h->length + 1;
  h->position_penalty_infinite = '\1';
  h->position_penalty = 0.0;
  
  if (OT_LEFT == o_type) {
    if (three_prime_base <= target_end) {
      h->position_penalty_infinite = '\0';
      if (three_prime_base < target_begin)
	h->position_penalty = target_begin - three_prime_base - 1;
      else {
	h->position_penalty = three_prime_base - target_begin + 1;
	inside_flag = 1;
      }
    }
  } else { /* OT_RIGHT == o_type */
    if (three_prime_base >= target_begin) {
      h->position_penalty_infinite = '\0';
      if (three_prime_base > target_end) {
	h->position_penalty = three_prime_base - target_end - 1;
      } else {
	h->position_penalty = target_end - three_prime_base + 1;
	inside_flag = 1;
      }
    }
  }
  if (!inside_flag)
    h->position_penalty *= pa->outside_penalty;
  else {
    h->position_penalty *= pa->inside_penalty;
  }
}

/* 
 * Return 1 if 'pair' spans any target (in sa->tar); otherwise return 0; An
 * oligo pair spans a target, t, if the last base of the left primer is to
 * left of the last base of t and the first base of the right primer is to
 * the right of the first base of t.  Of course the primers must
 * still be in a legal position with respect to each other.
 */
static int
pair_spans_target(pair, sa)
    const primer_pair *pair;
    const seq_args *sa;
{
    int i;
    int last_of_left = pair->left->start + pair->left->length - 1;
    int first_of_right = pair->right->start - pair->right->length + 1;
    int target_first, target_last;
    for (i = 0; i < sa->num_targets; i++) {
      target_first = sa->tar[i][0];
      target_last = target_first + sa->tar[i][1] - 1;
	if (last_of_left <= target_last
	    && first_of_right >= target_first
	    && last_of_left < first_of_right)
	    return 1;
    }
    return 0;
}

/* 
 * Return the value of the objective function value for given primer pair.
 * We must know that the pair is acceptable before calling obj_fn.
 */
static double
obj_fn(pa, h, int_num)
    const primer_args *pa;
    primer_pair *h;
    int  int_num;
{
    return h->left->quality + h->right->quality;
}

char *
pr_gather_warnings(sa, pa, ma)
    const seq_args *sa;
    const primer_args *pa, *ma;
{
    pr_append_str warning;

    PR_ASSERT(NULL != sa);
    PR_ASSERT(NULL != pa);
    PR_ASSERT(NULL != ma);

    warning.data = NULL;
    warning.storage_size = 0;

    if (pa->repeat_lib.warning.data)
	pr_append_new_chunk(&warning, pa->repeat_lib.warning.data);

    if(ma->repeat_lib.warning.data != NULL) {
	pr_append_new_chunk(&warning, ma->repeat_lib.warning.data);
	pr_append(&warning, " (for internal oligo)");
    }

    if (sa->warning.data) pr_append_new_chunk(&warning, sa->warning.data);
    return pr_is_empty(&warning) ? NULL : warning.data;
}

/* Print the data for chosen primer pairs to stdout in "boulderio" format. */
static void
boulder_print_pairs(prog_args, pa, ma, sa, best_pairs)
    const program_args *prog_args;
    const primer_args *pa, *ma;
    const seq_args *sa;
    const pair_array_t *best_pairs;
{
    const char *left_tag, *right_tag, *intl_tag, *prod_size_tag;
    char *warning;
    char suffix [3];
    primer_rec *fwd, *rev, *intl;
    int i, incl_s = sa->incl_s;

    PR_ASSERT(NULL != pa);
    PR_ASSERT(NULL != sa);
    PR_ASSERT(NULL != prog_args);

    if (prog_args->twox_compat) {
	left_tag = "FORWARD_PRIMER";
	right_tag = "REVERSE_PRIMER";
	intl_tag = "MIDDLE_OLIGO";
	prod_size_tag = "PRODUCT_SIZE";
    } else {
	left_tag = "PRIMER_LEFT";
	right_tag = "PRIMER_RIGHT";
	intl_tag = "PRIMER_INTERNAL_OLIGO";
	prod_size_tag = "PRIMER_PRODUCT_SIZE";
    }

    if ((warning = pr_gather_warnings(sa, pa, ma)) != NULL) {
	printf("PRIMER_WARNING=%s\n", warning);
	free(warning);
    }
    
    if (sa->error.data != NULL) {
	printf("PRIMER_ERROR=%s\n=\n", sa->error.data);
	if (fflush(stdout) == EOF) {
	    perror("fflush(stdout) failed");
	    exit(-1);
	}
	return;
    }

    if (pa->explain_flag) print_all_explain(pa, sa);

    for(i=0; i<best_pairs->num_pairs; i++) {
	fwd = best_pairs->pairs[i].left;
	rev = best_pairs->pairs[i].right;
	intl = best_pairs->pairs[i].intl;

	if (i == 0) suffix[0] = '\0';
	else sprintf(suffix, "_%d", i);

	printf("PRIMER_PAIR_QUALITY%s=%.4f\n", suffix,
			       best_pairs->pairs[i].pair_quality);

        /* Print sequences. */
	printf("PRIMER_LEFT%s_SEQUENCE=%s\n", suffix,
	       pr_oligo_sequence(sa, fwd));
	printf("PRIMER_RIGHT%s_SEQUENCE=%s\n", suffix,
	       pr_oligo_rev_c_sequence(sa, rev));
	if(pa->pick_internal_oligo)
#if USE_OLD_FORMAT_MISTAKE
	    printf("PRIMER_INTERNAL%s_OLIGO_SEQUENCE=%s\n", suffix,
#else
	    printf("PRIMER_INTERNAL_OLIGO%s_SEQUENCE=%s\n", suffix,
#endif
		   pr_oligo_sequence(sa,intl));

	printf("%s%s=%d,%d\n", left_tag, suffix,
	       fwd->start + incl_s + pa->first_base_index,
	       fwd->length);
	printf("%s%s=%d,%d\n", right_tag, suffix,
	       rev->start + incl_s + pa->first_base_index,
	       rev->length);
	if (pa->pick_internal_oligo)
	    printf("%s%s=%d,%d\n", intl_tag, suffix,
		   intl->start + incl_s + pa->first_base_index,
		   intl->length);

	printf("PRIMER_LEFT%s_TM=%.3f\n", suffix, fwd->temp);
	printf("PRIMER_RIGHT%s_TM=%.3f\n", suffix, rev->temp);
	if (pa->pick_internal_oligo)
	    printf("PRIMER_INTERNAL_OLIGO%s_TM=%.3f\n",suffix, intl->temp);
	printf("PRIMER_LEFT%s_SELF_ANY=%.2f\n", suffix,
	       fwd->self_any / PR_ALIGN_SCORE_PRECISION);
	printf("PRIMER_RIGHT%s_SELF_ANY=%.2f\n", suffix,
	       rev->self_any / PR_ALIGN_SCORE_PRECISION);
	if (pa->pick_internal_oligo)
	    printf("PRIMER_INTERNAL_OLIGO%s_SELF_ANY=%.2f\n", suffix,
		   intl->self_any / PR_ALIGN_SCORE_PRECISION);
	printf("PRIMER_LEFT%s_SELF_END=%.2f\n", suffix,
	       fwd->self_end / PR_ALIGN_SCORE_PRECISION);
	printf("PRIMER_RIGHT%s_SELF_END=%.2f\n",
	       suffix,rev->self_end / PR_ALIGN_SCORE_PRECISION);
	if (pa->pick_internal_oligo)
	    printf("PRIMER_INTERNAL_OLIGO%s_SELF_END=%.2f\n", suffix,
		   intl->self_end / PR_ALIGN_SCORE_PRECISION);
        if (pa->repeat_lib.seq_num > 0) {
	    printf("PRIMER_LEFT%s_MISPRIMING_SCORE=%.2f, %s\n", suffix,
		   fwd->repeat_sim.score[fwd->repeat_sim.max] / PR_ALIGN_SCORE_PRECISION,
		   fwd->repeat_sim.name);
            printf("PRIMER_RIGHT%s_MISPRIMING_SCORE=%.2f, %s\n", suffix,
		   rev->repeat_sim.score[rev->repeat_sim.max] / PR_ALIGN_SCORE_PRECISION,
                   rev->repeat_sim.name);
            printf("PRIMER_PAIR%s_MISPRIMING_SCORE=%.2f, %s\n", suffix,
		   best_pairs->pairs[i].repeat_sim / PR_ALIGN_SCORE_PRECISION,
		   best_pairs->pairs[i].rep_name);
        }
	if (pa->pick_internal_oligo && ma->repeat_lib.seq_num > 0) 
	    printf("PRIMER_INTERNAL_OLIGO%s_MISHYB_SCORE=%.2f, %s\n", 
		   suffix, intl->repeat_sim.score[intl->repeat_sim.max] / PR_ALIGN_SCORE_PRECISION,
		   intl->repeat_sim.name);
	if (pa->min_quality > 0){
	   printf("PRIMER_LEFT%s_MIN_SEQ_QUALITY=%d\n", suffix,
		   fwd->seq_quality);
           printf("PRIMER_RIGHT%s_MIN_SEQ_QUALITY=%d\n", suffix,
		   rev->seq_quality);
        }

	if (PR_DEFAULT_INSIDE_PENALTY != pa->inside_penalty
	    || PR_DEFAULT_OUTSIDE_PENALTY != pa->outside_penalty) {
	  printf("PRIMER_LEFT%s_POSITION_PENALTY=%f\n", suffix,
		 fwd->position_penalty);
	  printf("PRIMER_RIGHT%s_POSITION_PENALTY=%f\n", suffix,
		 rev->position_penalty);
#if 0
	    /* Really just debugging output. */
 	    printf("PRIMER_LEFT%s_QUALITY=%f\n", suffix,
		   fwd->quality);
   	    printf("PRIMER_RIGHT%s_QUALITY=%f\n", suffix,
		   rev->quality);
#endif
	}

	printf("PRIMER_LEFT%s_END_STABILITY=%.4f\n",
	       suffix, fwd->end_stability);
	printf("PRIMER_RIGHT%s_END_STABILITY=%.4f\n",
	       suffix, rev->end_stability);

	if (pa->pick_internal_oligo && ma->min_quality > 0)
	   printf("PRIMER_INTERNAL_OLIGO%s_MIN_SEQ_QUALITY=%d\n",
		   suffix, intl->seq_quality);
	printf("PRIMER_PAIR%s_COMPL_ANY=%.2f\n", suffix,
	       best_pairs->pairs[i].compl_any / PR_ALIGN_SCORE_PRECISION);
	printf("PRIMER_PAIR%s_COMPL_END=%.2f\n", suffix,
	       best_pairs->pairs[i].compl_end  / PR_ALIGN_SCORE_PRECISION);

	printf("%s%s=%d\n", prod_size_tag, suffix,
				     best_pairs->pairs[i].product_size);

    }
    printf("=\n");
    if (fflush(stdout) == EOF) {
	perror("fflush(stdout) failed");
	exit(-1);
    }
}

static void
print_all_explain(pa, sa)
    const primer_args *pa;
    const seq_args *sa;
{
    if(pa->explain_flag){
        print_explain(&sa->left_expl,OT_LEFT);
        print_explain(&sa->right_expl,OT_RIGHT);
        if (pa->pick_internal_oligo) print_explain(&sa->intl_expl, OT_INTL);
	printf("PRIMER_PAIR_EXPLAIN=");
	pr_print_pair_explain(stdout, sa);
    }
}

void
pr_print_pair_explain(f, sa)
    FILE *f;
    const seq_args *sa;
{
    fprintf(f, "considered %d",sa->pair_expl.considered);
    if (sa->pair_expl.target)
	fprintf(f, ", no target %d", sa->pair_expl.target);
    if (sa->pair_expl.product)
	fprintf(f, ", unacceptable product size %d", sa->pair_expl.product);
    if (sa->pair_expl.temp_diff) 
	fprintf(f, ", tm diff too large %d",sa->pair_expl.temp_diff);
    if (sa->pair_expl.compl_any) 
	fprintf(f, ", high any compl %d", sa->pair_expl.compl_any);
    if (sa->pair_expl.compl_end) 
	fprintf(f, ", high end compl %d", sa->pair_expl.compl_end);
    if (sa->pair_expl.internal) 
	fprintf(f, ", no internal oligo %d", sa->pair_expl.internal);
    if (sa->pair_expl.repeat_sim)
	fprintf(f, ", high mispriming library similarity %d",
				   sa->pair_expl.repeat_sim);
    fprintf(f, ", ok %d\n", sa->pair_expl.ok);
}

static void
print_explain(stat, l)
    const oligo_stats *stat;
    oligo_type l;
{
    if(OT_LEFT == l)printf("PRIMER_LEFT_EXPLAIN=");
    else if(OT_RIGHT == l)printf("PRIMER_RIGHT_EXPLAIN=");
	 else printf("PRIMER_INTERNAL_OLIGO_EXPLAIN=");

    printf("considered %d", stat->considered);
    if(stat->ns)printf(", too many Ns %d", stat->ns);
    if(stat->target)printf(", overlap target %d", stat->target);
    if(stat->excluded)printf(", overlap excluded region %d", stat->excluded);
    if(stat->gc)printf(", GC content failed %d", stat->gc);
    if(stat->gc_clamp)printf(", GC clamp failed %d", stat->gc_clamp);
    if(stat->temp_min)printf(", low tm %d", stat->temp_min);
    if(stat->temp_max)printf(", high tm %d", stat->temp_max);
    if(stat->compl_any)printf(", high any compl %d", stat->compl_any);
    if(stat->compl_end)printf(", high end compl %d", stat->compl_end);
    if(stat->repeat)printf(", high repeat similarity %d", stat->repeat);
    if(stat->poly_x)printf(", long poly-x seq %d", stat->poly_x);
    if(stat->seq_quality)printf(",low sequence quality %d", stat->seq_quality);
    if (stat->stability) printf(",high 3' stability %d", stat->stability);
    printf(", ok %d\n", stat->ok);
}

/*
 * Return 1 on error, 0 on success.  Set sa->trimmed_seq and possibly modify
 * sa->tar.  Upcase and check all bases in sa->trimmed_seq.
 */
static int
data_control(pa, ma, sa)
    primer_args *pa, *ma;
    seq_args *sa;
{
    static char s1[MAX_PRIMER_LENGTH+1];
    int i, pr_min;
    int seq_len = strlen(sa->sequence);
    int liberal_base_warning = 0;

    if (pa->primer_max_size > MAX_PRIMER_LENGTH) {
	pr_append_new_chunk(&pa->glob_err,
			    "PRIMER_MAX_SIZE exceeds built-in maximum");
	return 1;
    }

    if (pa->primer_opt_size > pa->primer_max_size) {
	pr_append_new_chunk(&pa->glob_err,
			    "PRIMER_{OPT,DEFAULT}_SIZE > PRIMER_MAX_SIZE");
	return 1;
    }

    if (pa->primer_opt_size < pa->primer_min_size) {
	pr_append_new_chunk(&pa->glob_err,
			    "PRIMER_{OPT,DEFAULT}_SIZE < PRIMER_MIN_SIZE");
	return 1;
    }

    if (ma->primer_max_size > MAX_PRIMER_LENGTH) {
	pr_append_new_chunk(&pa->glob_err,
		  "PRIMER_INTERNAL_OLIGO_MAX_SIZE exceeds built-in maximum");
        return 1;
    }

    if (ma->primer_opt_size > ma->primer_max_size) {
	pr_append_new_chunk(&pa->glob_err,
		  "PRIMER_INTERNAL_OLIGO_{OPT,DEFAULT}_SIZE > MAX_SIZE");
        return 1;
    }

    if (ma->primer_opt_size < ma->primer_min_size) {
	pr_append_new_chunk(&pa->glob_err,
		  "PRIMER_INTERNAL_OLIGO_{OPT,DEFAULT}_SIZE < MIN_SIZE");
        return 1;
    }

    if (pa->gc_clamp > pa->primer_min_size) {
	pr_append_new_chunk(&pa->glob_err,
			    "PRIMER_GC_CLAMP > PRIMER_MIN_SIZE");
	return 1;
    }

    if (NULL == sa->sequence_name && pa->file_flag) {
	pr_append_new_chunk(&sa->error,
			    "Need PRIMER_SEQUENCE_ID if PRIMER_FILE_FLAG != 0");
	return 1;
    }

    if (0 == pa->num_intervals) {
	pr_append_new_chunk(&pa->glob_err,
			    "Empty value for PRIMER_PRODUCT_SIZE_RANGE");
	return 1;
    }
    for (i = 0; i < pa->num_intervals; i++) {
	if (pa->pr_min[i] > pa->pr_max[i] || pa->pr_min[i] < 0) {
	    pr_append_new_chunk(&pa->glob_err,
				"Illegal element in PRIMER_PRODUCT_SIZE_RANGE");
	    return 1;
	}
    }

    pr_min = INT_MAX;
    for(i=0;i<pa->num_intervals;i++)
	if(pa->pr_min[i]<pr_min) pr_min=pa->pr_min[i];

    if (pa->primer_max_size > pr_min) {
	pr_append_new_chunk(&pa->glob_err,
			    "PRIMER_MAX_SIZE > min PRIMER_PRODUCT_SIZE_RANGE");
	return 1;
    }

    if (ma->primer_max_size > pr_min) {
	pr_append_new_chunk(&pa->glob_err,
		 "PRIMER_INTERNAL_OLIGO_MAX_SIZE > min PRIMER_PRODUCT_SIZE_RANGE");
        return 1;
    }

    if (pa->num_return < 1) {
	pr_append_new_chunk(&pa->glob_err,
			    "PRIMER_NUM_RETURN < 1");
        return 1;
    }

    if (sa->incl_l >= INT_MAX) {
	pr_append_new_chunk(&sa->error, "Value for INCLUDED_REGION too large");
	return 1;
    }

    if (sa->incl_s < 0 || sa->incl_l < 0 
	|| sa->incl_s + sa->incl_l > seq_len) {
	pr_append_new_chunk(&sa->error, "Illegal value for INCLUDED_REGION");
	return 1;
    }
    
    if (sa->incl_l < pr_min) {
	pr_append_new_chunk(&sa->error,
			    "INCLUDED_REGION length < min PRIMER_PRODUCT_SIZE_RANGE");
	return 1;
    }

    if (pa->max_end_stability < 0) {
        pr_append_new_chunk(&sa->error,
			    "PRIMER_MAX_END_STABILITY must be non-negative");
	return 1;
    }

    sa->trimmed_seq = pr_safe_malloc(sa->incl_l + 1);
    substr(sa->sequence, sa->incl_s, sa->incl_l, sa->trimmed_seq);

    if (check_intervals("TARGET", sa->num_targets, sa->tar, seq_len, sa)
	== 1) return 1;

    if (check_intervals("EXCLUDED_REGION", sa->num_excl, sa->excl,
			seq_len, sa)
	== 1) return 1;

    if (check_intervals("PRIMER_INTERNAL_OLIGO_EXCLUDED_REGION",
			sa->num_internal_excl, sa->excl_internal,
			seq_len, sa)
	== 1) return 1;

    if (pa->min_quality != 0 || ma->min_quality != 0){
	if(pa->min_quality != 0 && pa->min_quality < pa->quality_range_min) {
	   pr_append_new_chunk(&pa->glob_err,
	       "PRIMER_MIN_QUALITY < PRIMER_QUALITY_RANGE_MIN");
           return 1;
        }
	if(pa->min_quality != 0 && pa->min_quality > pa->quality_range_max) {
	   pr_append_new_chunk(&pa->glob_err,
	       "PRIMER_MIN_QUALITY > PRIMER_QUALITY_RANGE_MAX");
           return 1;
        }
	if(ma->min_quality != 0 && ma->min_quality <pa->quality_range_min) {
	   pr_append_new_chunk(&pa->glob_err,
	    "PRIMER_INTERNAL_OLIGO_MIN_QUALITY < PRIMER_QUALITY_RANGE_MIN");
           return 1;
        }
	if(ma->min_quality != 0 && ma->min_quality > pa->quality_range_max) {
	   pr_append_new_chunk(&pa->glob_err,
	     "PRIMER_INTERNAL_OLIGO_MIN_QUALITY > PRIMER_QUALITY_RANGE_MAX");
           return 1;
        }
	for(i=0; i< seq_len; i++) {
	   if(sa->quality[i] < pa->quality_range_min ||
		 sa->quality[i] > pa->quality_range_max) {
             pr_append_new_chunk(&sa->error,
		"Sequence quality score out of range");
             return 1;
           }
        }
    }

    seq_len = TRIMMED_SEQ_LEN(sa);
    for(i = 0; i < seq_len; i++){
	switch (sa->trimmed_seq[i])
	{
	case 'a': sa->trimmed_seq[i]='A'; break;
	case 'c': sa->trimmed_seq[i]='C'; break;
	case 'g': sa->trimmed_seq[i]='G'; break;
	case 't': sa->trimmed_seq[i]='T'; break;
	case 'n': sa->trimmed_seq[i]='N'; break;
	case 'A': break;
	case 'C': break;
	case 'G': break;
	case 'T': break;
	case 'N': break;
	default: 
	    if (pa->liberal_base) {
		sa->trimmed_seq[i] = 'N';
		liberal_base_warning = 1;
	    }
	    else {
		pr_append_new_chunk(&sa->error, "Unrecognized base in input sequence");
		return 1;
	    }
	}
    }
    if (liberal_base_warning)
	pr_append_new_chunk(&sa->warning,
			    "Unrecognized base in input sequence");

    if(pa->opt_tm < pa->min_tm || pa->opt_tm > pa->max_tm) {
	 pr_append_new_chunk(&pa->glob_err,
			     "Illegal values for PRIMER_TM");
	 return 1;
    }
    if(ma->opt_tm < ma->min_tm || ma->opt_tm > ma->max_tm) {
	 pr_append_new_chunk(&pa->glob_err,
			     "Illegal values for PRIMER_INTERNAL_OLIGO_TM");
	 return 1;
    }
    if(pa->min_gc>pa->max_gc||pa->min_gc>100||pa->max_gc<0){
	 pr_append_new_chunk(&pa->glob_err,
			     "Illegal value for PRIMER_MAX_GC and PRIMER_MIN_GC");
	 return 1;
    }
    if(ma->min_gc>ma->max_gc||ma->min_gc>100||ma->max_gc<0){
	 pr_append_new_chunk(&pa->glob_err,
			     "Illegal value for PRIMER_INTERNAL_OLIGO_GC");
	 return 1;
    }
    if(pa->num_ns_accepted<0){
	 pr_append_new_chunk(&pa->glob_err,
			     "Illegal value for PRIMER_NUM_NS_ACCEPTED");
	 return 1;
    }
    if(ma->num_ns_accepted<0){
	 pr_append_new_chunk(&pa->glob_err,
			     "Illegal value for PRIMER_INTERNAL_OLIGO_NUM_NS");
	 return 1;
    }
    if(pa->self_any<0||pa->self_end<0
		       ||pa->pair_compl_any<0||pa->pair_compl_end<0){
         pr_append_new_chunk(&pa->glob_err,
	     "Illegal value for primer complementarity restrictions");
	 return 1;
    }
    if(ma->self_any<0||ma->self_end<0){
	  pr_append_new_chunk(&pa->glob_err,
	      "Illegal value for internal oligo complementarity restrictions");
	  return 1;
    }
    if(pa->salt_conc<=0||pa->dna_conc<=0){
	  pr_append_new_chunk(&pa->glob_err,
	      "Illegal value for primer salt or dna concentration");
	  return 1;
    }
    if(ma->salt_conc<=0||ma->dna_conc<=0){
	  pr_append_new_chunk(&pa->glob_err,
	      "Illegal value for internal oligo salt or dna concentration");
	  return 1;
    }
    if (!DEFAULT_POSITION_PENALTIES(pa)	&& sa->num_targets > 1) {
      pr_append_new_chunk(&sa->error,
			  "Non-default inside penalty or outside penalty ");
      pr_append(&sa->error,
		"is valid only when number of targets <= 1");
    }
    if (!DEFAULT_POSITION_PENALTIES(pa)	&& 0 == sa->num_targets) {
      pr_append_new_chunk(&sa->warning,
			  "Non-default inside penalty or outside penalty ");
      pr_append(&sa->warning,
		"has no effect when number of targets is 0");
    }
    if (!pa->pick_internal_oligo && sa->internal_input) {
      pr_append_new_chunk(&sa->error,
			  "Not specified to pick internal oligos");
      pr_append(&sa->error,
		" but a specific internal oligo is provided");
    }
    if (sa->internal_input) {
      if (strlen(sa->internal_input) > ma->primer_max_size)
	pr_append_new_chunk(&sa->error, "Specified internal oligo too long");
      if (!strstr(sa->sequence, sa->internal_input))
	pr_append_new_chunk(&sa->error,
			    "Specified internal oligo not in sequence");
    }
    if (sa->left_input) {
      if (strlen(sa->left_input) > pa->primer_max_size)
	pr_append_new_chunk(&sa->error, "Specified left primer too long");
      if (!strstr(sa->sequence, sa->left_input))
	pr_append_new_chunk(&sa->error,
			    "Specified left primer not in sequence");
    }
    if (sa->right_input) {
      if (strlen(sa->right_input) > pa->primer_max_size) {
	pr_append_new_chunk(&sa->error, "Specified right primer too long");
       } else { /* We do not want to overflow s1. */
	reverse_complement(sa->right_input,s1);
	if (!strstr(sa->sequence, s1))
	  pr_append_new_chunk(&sa->error,
			      "Specified right primer not in sequence");
      }
    }
    if (pa->file_flag) pa->compl_flag = 1;
    return NULL == sa->error.data ? 0 : 1;
}

/* 
 * Check intervals, and add to sa->error.  Update the start of each interval to
 * be relative to the start of the included region.
 */ 
static int
check_intervals(tag_name, num_intervals, intervals, seq_len, sa)
    const char *tag_name;
    const int num_intervals;
    interval_array_t intervals;
    const int seq_len;
    seq_args *sa;
{
    int i;
    int outside_warning_issued = 0;
    for (i=0; i < num_intervals; i++) {
	if (intervals[i][0] + intervals[i][1] > seq_len) {
	    pr_append_new_chunk(&sa->error, tag_name);
	    pr_append(&sa->error, " beyond end of sequence");
	    return 1;
	}
	/* Cause the interval start to be relative to the included region. */
	intervals[i][0] -= sa->incl_s;
	/* Check that intervals are within the included region. */
	if (intervals[i][0] < 0
	    || intervals[i][0] + intervals[i][1] > sa->incl_l) {
	    if (!outside_warning_issued) {
		pr_append_new_chunk(&sa->warning, tag_name);
		pr_append(&sa->warning,
			  " outside of INCLUDED_REGION");
		outside_warning_issued = 1;
	    }
	}
	if (intervals[i][1] < 0) {
	    pr_append_new_chunk(&sa->error, "Negative ");
	    pr_append(&sa->error, tag_name);
	    pr_append(&sa->error, " length");
	    return 1;
	}
    }
    return 0;
}

void
pr_append(x, s)
    pr_append_str *x;
    const char *s;
{
    int xlen, slen;
    if (NULL == x->data) {
	x->storage_size = 24;
	x->data = pr_safe_malloc(x->storage_size);
	*x->data = '\0';
    }
    xlen = strlen(x->data);
    slen = strlen(s);
    if (xlen + slen + 1 > x->storage_size) {
	x->storage_size += 2 * (slen + 1);
	x->data = pr_safe_realloc(x->data, x->storage_size);
    }
    strcpy(x->data + xlen, s);
}

void
pr_append_new_chunk(x, s)
    pr_append_str *x;
    const char *s;
{
    if (pr_is_empty(x))
	pr_append(x, s);
    else {
	pr_append(x, "; ");
	pr_append(x, s);
    }
}

void
pr_set_empty(x)
    pr_append_str *x;
{
    PR_ASSERT(NULL != x);
    if (NULL != x->data) *x->data = '\0';
}

int
pr_is_empty(x)
    const pr_append_str *x;
{
    PR_ASSERT(NULL != x);
    return  NULL == x->data || '\0' == *x->data;
}

static short
align(s1, s2, a)
    const char *s1, *s2;
    const dpal_args *a;
{
    dpal_results r;
    dpal(s1, s2, a, &r);
    PR_ASSERT(r.score <= SHRT_MAX);
    return ((r.score<0) ? 0 : (short)r.score);
}

/* Set dpal args to appropriate values for primer picking. */
static void
set_dpal_args(a)
    dpal_args *a;
{
    unsigned int i, j;

    memset(a, 0, sizeof(*a));
    for (i = 0; i <= UCHAR_MAX; i++)
	for (j = 0; j <= UCHAR_MAX; j++)
	    if (('A' == i || 'C' == i || 'G' == i || 'T' == i || 'N' == i)
		&& ('A' == j || 'C' == j || 'G' == j || 'T' == j 
		    || 'N' == j)) {
		    if (i == 'N' || j == 'N') 
			a->ssm[i][j] = -25;
		    else if (i == j)
			a->ssm[i][j] = 100;
		    else 
			a->ssm[i][j] = -100;
		} else
		    a->ssm[i][j] = INT_MIN;

    a->gap                = -200;
    a->gapl               = -200;
    a->flag               = DPAL_LOCAL;
    a->max_gap            = 1;
    a->fail_stop          = 1;
    a->check_chars        = 0;
    a->debug              = 0;
    a->score_only         = 1;
    a->force_generic      = 0;
    a->force_long_generic = 0;
    a->force_long_maxgap1 = 0;
}

char *
pr_oligo_sequence(sa, o)
    const seq_args *sa;
    const primer_rec *o;
{
    static char s[MAX_PRIMER_LENGTH+1];
    int seq_len;
    PR_ASSERT(NULL != sa);
    PR_ASSERT(NULL != o);
    seq_len = strlen(sa->sequence);
    PR_ASSERT(o->start + sa->incl_s >= 0);
    PR_ASSERT(o->start + sa->incl_s + o->length <= seq_len);
    substr(sa->sequence, sa->incl_s + o->start, o->length, s);
    return &s[0];
}

char *
pr_oligo_rev_c_sequence(sa, o)
    const seq_args *sa;
    const primer_rec *o;
{
    static char s[MAX_PRIMER_LENGTH+1], s1[MAX_PRIMER_LENGTH+1];
    int seq_len, start;
    PR_ASSERT(NULL != sa);
    PR_ASSERT(NULL != o);
    seq_len = strlen(sa->sequence);
    start = sa->incl_s + o->start - o->length + 1;
    PR_ASSERT(start >= 0);
    PR_ASSERT(start + o->length <= seq_len);
    substr(sa->sequence, start, o->length, s);
    reverse_complement(s,s1);
    return &s1[0];
}

static void
print_usage()
{
    const char **p;
    p = &copyright[0];
    while (NULL != *p) fprintf(stderr, "%s\n", *p++);
    fprintf(stderr, 
	    "\n\nUSAGE: %s %s %s\n", pr_program_name,
	    "[-format_output] [-2x_compat]",
	    "[-strict_tags]");
    fprintf(stderr, "This is primer3 (%s)\n", pr_release);
    fprintf(stderr, "Input must be provided on standard input.\n");
}

static void
sig_handler(signal)
    int signal;
{
    print_usage();
    fprintf(stderr, "%s: received signal %d\n", pr_program_name, signal);
    exit(signal);
}

static void
oligo_compl(h, ha, sa, l, local_args, end_args, local_end_args)
    primer_rec *h;
    const primer_args *ha;
    seq_args *sa;
    oligo_type l;
    const dpal_args *local_args, *end_args, *local_end_args;
{
    int i, j, min, max;
    char s[MAX_PRIMER_LENGTH+1], s1[MAX_PRIMER_LENGTH+1];
    double w;

    h->ok = 0;
    j =  (OT_LEFT == l || OT_INTL == l) ? h->start : h->start-h->length+1;

    substr(sa->trimmed_seq, j, h->length, s1);
    reverse_complement(s1, s);

    h->self_any = align(s1, s, local_args);
    if(h->self_any > ha->self_any){
	h->ok = 6;
	if      (OT_LEFT  == l) {
	     sa->left_expl.compl_any++;
	     sa->left_expl.ok--;
        }
	else if (OT_RIGHT == l) {
	     sa->right_expl.compl_any++;
	     sa->right_expl.ok--;
        }
	else {
	     sa->intl_expl.compl_any++;
	     sa->intl_expl.ok--;
        }
	return;
    }

    h->self_end = (l != OT_RIGHT) ? align(s1,s,end_args) : align(s,s1,end_args);
    if(h->self_end > ha->self_end) {
	h->ok = 7;
	if      (OT_LEFT  == l) {
	    sa->left_expl.compl_end++;
	    sa->left_expl.ok--;
        }
	else if (OT_RIGHT == l) {
	    sa->right_expl.compl_end++;
	    sa->right_expl.ok--;
        }
	else {
	    sa->intl_expl.compl_end++;
	    sa->intl_expl.ok--;
        }
	return;
    }

    /*
     * Calculate maximum similarity to sequences from user defined repeat
     * library. Compare it with maximum allowed repeat similarity.
     */
    if(ha->repeat_lib.seq_num > 0){
	if(h->repeat_sim.score != NULL) free(h->repeat_sim.score);
	h->repeat_sim.score = 
	  /* 
	   * leak here; h->repeat_sim.score is not freed when oligo lists
	   * are recycled.
	   */
	  pr_safe_malloc(ha->repeat_lib.seq_num*sizeof(short));
        h->repeat_sim.max = h->repeat_sim.min = 0;
	max = min = 0;
	h->repeat_sim.name = ha->repeat_lib.names[0];
	for(i = 0; i < ha->repeat_lib.seq_num; i++){
	  if (OT_LEFT == l) w = ha->repeat_lib.weight[i] *
	       align(s1,ha->repeat_lib.seqs[i], local_end_args);
          else if (OT_INTL == l) w = ha->repeat_lib.weight[i] *
	       align(s1,ha->repeat_lib.seqs[i], local_args);
          else w = ha->repeat_lib.weight[i] *
	       align(s, ha->repeat_lib.rev_compl_seqs[i], local_end_args);

	  h->repeat_sim.score[i] = w;
	  if(w > max){
	     max = w;
	     h->repeat_sim.max = i;
	     h->repeat_sim.name = ha->repeat_lib.names[i];
          }
	  if(w < min){
	     min = w;
	     h->repeat_sim.min = i;
          }
	  if(w > ha->repeat_compl) {
		  h->ok = 10;
		  if      (OT_LEFT  == l) {
		      sa->left_expl.repeat++;
		      sa->left_expl.ok--;
                  }
		  else if (OT_RIGHT == l) {
		      sa->right_expl.repeat++;
		      sa->right_expl.ok--;
                  }
		  else {
		      sa->intl_expl.repeat++;
		      sa->intl_expl.ok--;
                  }
		  return;
          }
	}

     } 

}

static int
pair_repeat_sim(h, pa)
    primer_pair *h;
    const primer_args *pa;
{
    int i, n, max, w;
    primer_rec *fw, *rev;
     
    fw = h->left;
    rev = h->right;

    max = 0;
    n = pa->repeat_lib.seq_num;
    if(n == 0) return 0;
    h->rep_name =  pa->repeat_lib.names[0] ;
    for(i = 0; i < n; i++){
	if((w=(fw->repeat_sim.score[i] +
		rev->repeat_sim.score[i])) > max) {
		    max = w;
		    h->rep_name =  pa->repeat_lib.names[i] ;
                }
    }
        return max;
}

/* =========================================================== */
/* Various fail-stop wrappers for standard library functions.  */

void *
pr_safe_malloc(x)
    size_t x;
{
    void *r = malloc(x);
    if (NULL == r) OOM_ERROR;
    return r;
}

void *
pr_safe_realloc(p, x)
    void *p;
    size_t x;
{
    void *r = realloc(p, x);
    if (NULL == r) OOM_ERROR;
    return r;
}

static FILE *
safe_fopen(path, mode)
    const char *path, *mode;
{
    FILE *r = fopen(path, mode);
    if (NULL == r) {
	fprintf(stderr, "%s: unable to open file %s:",
		pr_program_name, path);
	perror("");
	exit (-1);
    }
    return r;
}

/* =========================================================== */
